#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import persistent
from appy.px import Px
from appy.ui import Iframe
from appy.ui.layout import Layout
from appy.model.fields import Show
from appy.database.lock import Lock
from appy.ui.template import Template
from appy.ui.criteria import Criteria
from appy.model.fields.ref import Ref
from appy.ui.validate import Validator
from appy.model.searches import Search
from appy.model.fields.text import Text
from appy.model.fields.date import Date
from appy.model.utils import Object as O
from appy.database.catalog import Catalog
from appy.xml.marshaller import Marshaller
from appy.model.fields.string import String
from appy.model.fields.action import Action
from appy.model.fields.computed import Computed
from appy.model.workflow.history import History
from appy.model.fields.phase import Page, UiPhases
from appy.utils import multicall, MessageException
from appy.model.workflow import standard as workflows
from appy.model.fields.select import Select, Selection
from appy.utils.string import Normalize, produceNiceMessage

# Errors - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
RO_EXPR     = 'This is a read-only property: you cannot set a value for it.'
FIELD_KO    = 'Field "%s" on class "%s" does not exist.'
MISS_CLS    = '"new" called without class name.'
REF_INS_KO  = 'Param "insert" for Ref field %s is None.'
CONT_ERR    = "Container can't be retrieved that way."
INDEX_TEMP  = "A temp object can't be (un)indexed."
CATALOG_KO  = 'Catalog not found for %s instance.'
NO_ID       = 'Call to getObject with no ID.'
NO_STR      = 'Specify the name of the class as a string.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Base:
    '''Base class for any Appy class'''

    # Some elements will be traversed
    traverse = {'history': True}

    # Make some elements available here
    Lock = Lock

    # The default workflow for a class that does not define one
    workflow = workflows.Anonymous

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Fields
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # The following fields are base fields that apply to any class, be it a
    # class from the base Appy model or a class (re)defined in the app.

    # Field "searchable" defines an index containing keywords collected from all
    # other object fields having attribute searchable=True.
    def getSearchableText(self):
        '''Collects and returns keywords as defined on every field being
           "searchable".'''
        r = []
        for field in self.class_.fields.values():
            if not field.searchable: continue
            val = field.getIndexValue(self, searchable=True)
            if val:
                r.append(val)
        return ' '.join(r)

    searchable = Text(show=False, persist=False, indexed=True,
                      default=getSearchableText, label='Base', page=None)

    # Attribute "title" is a mandatory field used at various places by default,
    # for displaying base information about an object.
    titleAttributes = {'multiplicity': (1,1), 'show': Show.EX, 'indexed': True,
      'searchable': True, 'filterField': searchable,
      # This method allows to produce an index value suitable for sorting
      'indexValue': lambda o, v: Normalize.text(o.getShownValue(),
                                                keepBlank=False) if v else ''}

    title = String(**titleAttributes)

    # Attribute "creator" gets the login of the user having created the object
    p = {'multiplicity': (1,1), 'show': 'xml', 'indexed': True, 'label': 'Base'}
    creator = Computed(method=lambda o: o.history[-1].login, **p)

    def setCreator(self, login):
        '''Sets p_login as p_self's creator'''
        self.history[-1].login = login

    # Object's creation and last modification dates
    p['show'] = Show.VE_
    created = Date(default=lambda o: o.history[-1].date, persist=False, **p)
    modified = Date(default=lambda o: o.history.modified, persist=False, **p)

    # "state" is a field deduced from the object history and identifies the
    # object state according to its workflow. It must be a "select" field,
    # because it will be necessary for displaying the translated state name.
    state = Select(validator=Selection(lambda o: o.getWorkflow().listStates(o)),
              svalidator=Selection(lambda tool, c: c.workflow.listStates(tool)),
              show='result', default=lambda o: o.history[0].state,
              persist=False, indexed=True, height=5, label='Base')

    # Field "allowed" defines an index storing the list of roles and users being
    # allowed to view this object. It will be used within catalog searches with
    # attribute "secure" being "True", for filtering objects the user is allowed
    # to see.
    def getAllowed(self):
        '''Returns the list of roles and users which are granted read access to
           this object.'''
        allowedRoles = self.getWorkflow().getRolesFor(self, 'read')
        if self.localRoles.only:
            # The roles will not be mentioned as-is
            r = []
        else:
            r = list(allowedRoles.keys())
        # Add users or groups having, locally, this role on this object
        for id, roles in self.localRoles.items():
            for role in roles:
                if role in allowedRoles:
                    usr = 'user:%s' % id
                    if usr not in r: r.append(usr)
        return r

    allowed = Select(multiplicity=(0, None), show=False, persist=False,
                     indexed=True, default=getAllowed, label='Base')

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # History
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    recordAttributes = {'method': History.view, 'label':'Base', 'layouts':'f',
      'page':Page('history', label='Base_page_history',
                  sticky=True, icon='history.svg',
                  show=lambda o: 'view' if o.user.hasRole('Manager') else None)}

    record = Computed(**recordAttributes)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Properties
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # The following properties are read-only attributes allowing easy access,
    # from any object in the database, to (i) a series of heavily-demanded
    # objects, like the tool or the currently logged user, or (ii) commonly
    # used functions.
    # ~~~ All these properties will have this setter, raising an exception
    def raiseReadOnly(self, v): raise Exception(RO_EXPR)

    # The guard
    guard = property(lambda o: o.H().guard, raiseReadOnly)

    # The currently logged user
    user = property(lambda o: o.H().guard.user, raiseReadOnly)

    # The unique Tool instance
    tool = property(lambda o: o.H().tool, raiseReadOnly)

    # Get the appy.model.meta.Class for this instance
    class_ = property(lambda o: o.getClass(), raiseReadOnly)

    # Get the apps' configuration
    config = property(lambda o: o.H().config, raiseReadOnly)

    # The request and response objects
    req = property(lambda o: o.H().req, raiseReadOnly)
    resp = property(lambda o: o.H().resp, raiseReadOnly)

    # The object (base) URL
    url = property(lambda o: o.H().server.getUrl(o), raiseReadOnly)

    # A potential initiator object
    initiator = property(lambda o: o.getInitiator(), raiseReadOnly)

    # The referer URL
    referer = property(lambda o: o.H().headers['Referer'], raiseReadOnly)

    # The object container
    container = property(lambda o:o.getContainer(objectOnly=True),raiseReadOnly)

    # The app model
    model = property(lambda o: o.H().server.model, raiseReadOnly)

    # The site URL
    siteUrl = property(lambda o: o.config.server.getUrl(o.H()), raiseReadOnly)

    # The path to the app folder
    appPath = property(lambda o: o.H().config.model.appPath, raiseReadOnly)

    # The current traversal
    traversal = property(lambda o: o.H().traversal, raiseReadOnly)

    # The catalog storing indexed data for this object
    traverse['catalog'] = 'Manager'
    catalog = property(lambda o: o.getCatalog(), raiseReadOnly)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Base methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def getHandler(self):
        '''Returns the current request handler'''
        # Class "Handler" has been injected to class Base in Base.Handler in
        # appy.server.handler in order to avoid circular dependencies.
        return Base.Handler.get()

    # This method is used so often that it deserves a short name
    H = getHandler

    def getClass(self):
        '''Get the appy.model.meta.Class corresponding to p_self'''
        classes = self.H().server.model.classes
        name = self.__class__.__bases__[0].__name__
        return classes.get(name)

    def getWorkflow(self):
        '''Get the appy.model.meta.Workflow corresponding to p_self'''
        return self.getClass().workflow

    def do(self, name, comment=None, doAction=True, doHistory=True,
           secure=False, doSay=False, data=None):
        '''Programmatically triggers on p_self a transition named p_name'''
        # p_data can be a dict that will be included into the history event.
        # ~~~
        # If p_doSay is True, it will not lead to the UI message being
        # automatically included in the response cookie, but will allow the
        # caller to retrieve it (via the return value) and do what he wants with
        # it.
        transition = self.getWorkflow().transitions.get(name)
        return transition.trigger(self, comment=comment, doAction=doAction,
                     doHistory=doHistory, doSay=doSay, secure=secure, data=data)

    def getPageLayout(self, layout):
        '''Returns the page layout corresponding to p_layout for p_self'''
        # The layout may be forced by request parameter "pageLayout". Useful
        # when displaying an object in a particular context, ie, in a popup.
        req = self.req
        if 'pageLayout' in req:
            pl = req.pageLayout
            return pl if isinstance(pl, Layout) else Layout(pl)
        # Retrieve layouts being applicable for p_self
        r = self.class_.getPageLayouts(self)[layout]
        return r if not isinstance(r, str) else Layout(r)

    def getDefaultPage(self, layout):
        '''Get p_self's default page on p_layout'''
        # The defaut page for p_layout may be customized by special methods
        # m_getDefaultViewPage and m_getDefaultEditPage.
        method = 'getDefault%sPage' % layout.capitalize()
        return getattr(self, method)() if hasattr(self, method) else 'main'

    def getCurrentPage(self, req, layout):
        '''Gets the name of the currently shown page for p_self, or a default
           page if no page is defined in the p_req(uest).'''
        # Try to get the page from the request, or return p_self's default
        # page on p_layout.
        return req.page or self.getDefaultPage(layout)

    def getObject(self, id, logMissing=False):
        '''Gets an object given its p_id'''
        if not id: raise Exception(NO_ID)
        handler = self.H()
        return handler.server.database.getObject(handler, id,
                                                 logMissing=logMissing)

    def __repr__(self):
        '''Returns the class name and Appy object ID'''
        return '<%s id=%s>' % (self.getClass().name, self.id)

    def isTemp(self):
        '''Is this object temporary ?'''
        return self.id in self.H().dbConnection.root.temp

    def log(self, message, type='info', noUser=False):
        '''Logs a p_message of some p_type'''
        return self.H().log('app', type, message)

    def say(self, message):
        '''Adds p_message to the global message returned in the response'''
        self.resp.addMessage(message)

    def goto(self, url=None, message=None):
        '''Return to some p_url or to the referer page'''
        return self.resp.goto(url, message)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Security methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def raiseUnauthorized(self, msg=None):
        '''Raise a security-related error'''
        raise self.guard.Error(msg)

    def raiseMessage(self, message):
        '''Raise a nicely-rendered error, without technical details'''
        raise MessageException(message)

    def allows(self, permission='read', raiseError=False):
        '''Check the guard's homonym method'''
        return self.guard.allows(self, permission, raiseError=raiseError)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # URL methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def buildUrl(self, name='', base=None, ram=False, bg=False):
        '''See doc in appy.server.Server::buildUrl'''
        H = self.H()
        return H.server.buildUrl(H, name=name, base=base, ram=ram, bg=bg)

    def getUrl(self, sub=None, **params):
        '''See doc in appy.server.Server::getUrl'''
        return self.H().server.getUrl(self, sub=sub, **params)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Field methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def getField(self, name, className=None, raiseError=False):
        '''Returns the field whose name is p_name. If p_className is there, get
           the field on this class instead of on p_self's class. If p_raiseError
           is True and the field is not found, an exception is raised. Else,
           None is returned.'''
        class_ = self.H().server.model.classes[className] if className \
                                                          else self.getClass()
        r = class_.fields.get(name)
        if r is None and raiseError:
            raise AttributeError(FIELD_KO % (name, class_.name))
        return r

    def getFields(self, layout, page=None, phase=None, type=None, fields=None):
        '''Returns the list of fields to render on p_layout'''
        r = []
        # Browse p_self's fields, or p_fields if passed
        for field in (fields or self.class_.fields.values()):
            # Ignore fields not being on p_page
            if page and (field.pageName != page): continue
            # Ignore fields not being on any page in p_phase
            if phase and (field.page and (field.page.phase != phase)): continue
            # Ignore fields not being of p_type
            if type and (field.type != type): continue
            # Ignore fields that can't be rendered on p_layout
            if not field.isRenderable(layout): continue
            # Ignore unshowable fields 
            if not field.isShowable(self, layout): continue
            r.append(field)
        return r

    def getGroupedFields(self, page, layout, collect=True, fields=None):
        '''Returns p_self's fields, sorted by group, being part of this p_page,
           to be shown on this p_layout.'''
        # If p_collect is...
        # ----------------------------------------------------------------------
        #  False  | r_ is a list, each element being a field or a group.
        # ----------------------------------------------------------------------
        #  True   | r_ is a tuple (page, l, css, js, phases):
        #         |  - "page"   repeats the current p_page;
        #         |  - "l"      is the list produced when p_collect is False;
        #         |  - "css"    is a list of CSS files that must be included
        #         |             because used by the returned fields;
        #         |  - "js"     is a list of Javascript files that must be
        #         |             included because used by the returned fields;
        #         |  - "phases" is an ordered dict of the phases and pages that
        #         |             must currently be shown for p_self.
        # ----------------------------------------------------------------------
        # If p_fields are given, we use them instead of getting all fields
        # defined on p_self's class.
        # ----------------------------------------------------------------------
        r = []
        groups = {} # The already encountered groups (UiGroup instances)
        css = js = phases = None # The elements to collect when relevant
        # Browse fields
        config = self.config
        for field in (fields or self.class_.fields.values()):
            if collect and field.page:
                # Collect phases, for all fields, not just for fields that will
                # be selected here.
                if phases is None: phases = UiPhases(page, self, layout)
                phases.addField(field)
            # Ignore fields whose page is not p_page
            if field.pageName != page: continue
            # Ignore fields that must not be shown
            if not field.isShowable(self, layout): continue
            if collect:
                # Collect JS/CSS files
                if css is None: css = []
                field.getCss(layout, css)
                if js is None: js = []
                field.getJs(layout, js, config)
            # Insert the field directly in the result if it is not in a group
            if not field.group:
                r.append(field)
            else:
                # Insert it in the right group (if the group exists on p_layout)
                group = field.getGroup(layout)
                if not group:
                    r.append(field)
                else:
                    # Insert the UiGroup instance corresponding to field.group
                    uiGroup = group.insertInto(r, groups, field.page,
                                               self.class_.name)
                    uiGroup.addElement(field)
        # Finalize the data structure storing phases
        if phases: phases.finalize()
        # At this time, we may realize that p_page can't be shown on p_layout.
        # In that case, re-compute everything for the default page.
        if phases and phases.unshowableCurrentPage():
            return self.getGroupedFields(phases.defaultPageName, layout,
                                         collect=collect, fields=fields)
        # Return the appropriate result, depending on p_collect
        return r if not collect else page, r, css, js, phases

    def isEmpty(self, name, value=None, fromParam=False):
        '''Returns True if value of field p_name is considered to be empty
           (or p_value if p_fromParam is True).'''
        field = self.getField(name, raiseError=True)
        if not fromParam:
            value = field.getStoredValue(self, name)
        return field.isEmptyValue(self, value)

    def getLabel(self, name, field=True):
        '''Gets the translated label for field (if p_field is True) or workflow
           state or transition (if p_field is False) named p_name.'''
        # Translate the name of a field
        if field:
            field = self.getField(name, raiseError=True)
            return self.translate('labelId', field=field)
        # Translate the name of a state or transition
        return self.translate('%s_%s' % (self.getWorkflow().name, name))

    def getShownValue(self, name='title', layout='view', language=None):
        '''Call field.getShownValue on field named p_name'''
        field = self.getField(name)
        return field.getShownValue(self, field.getValue(self), layout,
                                   language=language)

    def getValue(self, name, type=None, layout='view', language=None,
                 value=None, bypassDefault=False):
        '''Gets the value of field named p_name on p_self'''
        # In most cases, use the property set on p_self to retrieve this value.
        # For example, for retrieving the value of field named "title" on
        # p_self, use:
        #
        #                          self.title
        #
        # Use m_getValue if you need some variant of the field value. If p_type
        # is:
        # ----------------------------------------------------------------------
        #    None     | The value as stored in the database will be returned
        #             | (or p_value if not None). If p_bypassDefault is True, in
        #             | the absence of a value, the system allowing to retrieve
        #             | a default value will be bypassed. Note that it has no
        #             | sense to specify a p_value if p_type is None. It is
        #             | useful when specifying one of the 2 following types, in
        #             | order to format an alternate value than the one stored
        #             | on p_self.
        # ----------------------------------------------------------------------
        # 'formatted' | The "formatted" value, as produced by methods
        #             | <Field>.getFormattedValue, will be returned. For
        #             | example, in the case of a Select field, the formatted
        #             | value is a nice translated name, the database value
        #             | being a short technical term. If this formatting implies
        #             | translating something, it will be done in p_language, or
        #             | in the UI language if not specified. Parameter p_layout
        #             | may also be used, some fields using alternate value
        #             | representations on different layouts. Parameter p_layout
        #             | must be understood as "display the field value as it
        #             | would appear on layout X".
        # ----------------------------------------------------------------------
        #   'shown'   | A variant of the "formatted" type, taking care of
        #             | multilingual values, as produced by methods
        #             | <Field>.getShownValue. For a multilingual value, the
        #             | "shown" value will be the part corresponding to the UI
        #             | language (or to p_language if specified).
        # ----------------------------------------------------------------------
        #  'catalog'  | The field value as stored in the database index, as
        #             | produced by methods <Field>.getCatalogValue.
        # ----------------------------------------------------------------------
        field = self.getField(name)
        # Get the catalog value
        if type == 'catalog': return field.getCatalogValue(self)
        # All other values are based on the value as stored in the database.
        # Get it, excepted if given in p_value.
        if value is None:
            method = 'getStoredValue' if bypassDefault else 'getValue'
            r = getattr(field, method)(self, name)
        else:
            r = value
        if type is None: return r
        # Get the formatted or shown
        method = getattr(field, 'get%sValue' % type.capitalize())
        return method(self, r, layout, language=language)

    def countRefs(self, name):
        '''Counts the nb of objects linked to this one via Ref field p_name'''
        tied = self.values.get(name)
        return 0 if tied is None else len(tied)

    def getIndexOf(self, name, tied, raiseError=True):
        '''See docstring in the homonym method on appy/model/fields/ref::Ref'''
        return self.getField(name).getIndexOf(self, tied, raiseError=raiseError)

    def getPageIndexOf(self, name, tied):
        '''See docstring in the homonym method on appy/model/fields/ref::Ref'''
        return self.getField(name).getPageIndexOf(self, tied)

    def link(self, name, o, secure=False, executeMethods=True, at=None):
        '''Links p_o (which can be a list of objects) to p_self via Ref field
           p_name.'''
        # For understanding the 3 last params, check Ref's m_linkObject's doc
        field = self.getField(name)
        return field.linkObject(self, o, secure=secure,
                                executeMethods=executeMethods, at=at)

    def unlink(self, name, o, secure=False, executeMethods=True):
        '''Unlinks p_o (which can be a list of objects) from p_self via Ref
           field p_name.'''
        # For understanding the 2 last params, check Ref's m_unlinkObject's doc
        field = self.getField(name)
        return field.unlinkObject(self, o, secure=secure,
                                  executeMethods=executeMethods)

    def sort(self, name, sortKey='title', reverse=False):
        '''Sorts referred elements linked to p_self via Ref field named p_name
           according to a given p_sortKey which can be:
           - an attribute set on referred objects ("title", by default);
           - a method that will be called on every tied object, will receive
             every such object as unique arg and will return a value that will
             represent its order among all tied objects. This return value will
             then be returned by the standard method given to the "key" param of
             the standard list.sort method;
           - None. If None, default sorting will occur, using the method stored
             in field.insert.
        '''
        # Get the objects tied to p_self via field p_name
        objects = getattr(self, name, None)
        if not objects: return
        if not sortKey:
            # Sort according to field.insert
            field = self.getField(name)
            insertMethod = field.insert
            if not insertMethod:
                raise Exception(REF_INS_KO % name)
            if not callable(insertMethod): insertMethod = insertMethod[1]
            keyMethod = lambda o: insertMethod(self, o)
        elif isinstance(sortKey, str):
            # Sort according to p_sortKey
            field = objects[0].getField(sortKey)
            keyMethod = lambda o: field.getSortValue(o)
        else:
            # Sort according to a custom method
            keyMethod = lambda o: sortKey(o)
        # Perform the sort
        objects.sort(key=keyMethod, reverse=reverse)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Containment
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # Attribute "cid" stores identifying information about the object being the
    # container of this one. "cid" stands for "*c*ontainer *id*". An object
    # being tied to another one via a Ref field with attribute "composite" being
    # True defines such a "container relationship" between these objects, one
    # object being the container of the other. In that case, the contained
    # object will have a "cid" storing info about the container object and the
    # Ref field. More precisely, if:
    # 
    #   object X has a composite Ref R storing object Y, with back Ref B,
    #
    # the container ID for Y (Y.cid) will be equal to "<X.iid>_<B.name>"
    #
    # An object may be top-level and have no container. Such an object is named
    # a "root object" and its "cid" will be None.
    # 
    # It is illegal to have the same object being the target of more that one
    # composite Ref.

    def computeCid(self):
        '''Computes the container ID ("cid") for p_self'''
        r = None
        for field in self.class_.fields.values():
            if (field.type == 'Ref') and field.isBack and field.back.composite:
                container = getattr(self, field.name, None)
                if container:
                    try:
                        iid = container.iid
                    except AttributeError:
                        iid = container[0].iid
                    r = '%d_%s' % (iid, field.name)
                    break
        # If no container has been found, the object is a root object
        return r

    p['show'] = False
    cid = String(default=computeCid, persist=False, emptyIndexValue=None, **p)

    def getContainer(self, objectOnly=False, orInitiator=True, forward=False):
        '''Returns p_self's container, as a tuple (containerObject, fieldName)
           or the container alone if p_objectOnly is True. If the object is
           root, it returns (None, None) or None. If the container is not found
           and p_orInitiator is True, the initiator (which is probably the
           future container), when present, is returned.'''
        # Manage temp objects, having no container ID
        if self.isTemp():
            # Get the initiator
            initiator = self.getInitiator()
            if not initiator or not orInitiator:
                return None if objectOnly else (None, None)
            if not objectOnly and not forward:
                # In that case we cannot return the name of the back reference,
                # because the initiator gives the name of the forward reference
                # instead.
                raise Exception(CONT_ERR)
            o = initiator.o
            return o if objectOnly else (o, initiator.field.name)
        # Get the container ID
        cid = self.cid
        if cid is None:
            # A root object: it has no container
            return None if objectOnly else (None, None)
        else:
            id, name = cid.rsplit('_', 1)
            container = self.getObject(id)
            if forward and not objectOnly:
                # Get the name of p_name's forward reference
                name = self.getField(name).back.name
            return container if objectOnly else (container, name)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Catalog-related methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def search(self, className=None, **kwargs):
        '''Performs a search in the database, for instances of class named
           p_className, or of self's class if not specified.'''
        if className and not isinstance(className, str): raise Exception(NO_STR)
        handler = self.H()
        database = handler.server.database
        className = className or self.class_.name
        return database.search(handler, className, **kwargs)

    def search1(self, className=None, **kwargs):
        '''Performs a search for a single object'''
        r = self.search(className, **kwargs)
        return r[0] if r else None

    def count(self, className=None, **kwargs):
        '''Identical to m_search abo, but returns the number of objects matching
           the search instead of returning the objects themselves.'''
        # Use this method instead of writing len(self.search(...))
        return len(self.search(className=className, ids=True, **kwargs))

    def compute(self, className, sortBy=None, context=None, expression=None,
                secure=False, max=None, **fields):
        '''Computes a result based on a search on objects of p_className'''
        # Like m_search and m_count above, this method performs a search on
        # objects of p_className; but in this case, instead of returning a list
        # of matching objects (like m_search) or counting elements (like
        # p_count), it evaluates, on every matching object (possibly restricted
        # to at most p_max objects), a Python p_expression (which may be an
        # expression or a statement), and returns, if needed, a result. The
        # result may be initialized through parameter p_context. p_expression is
        # evaluated with 2 variables in its context: "o" which is the currently
        # walked object, instance of p_className, and "ctx", which is the
        # context as initialized (or not) by p_context. p_context may be used as
        # (1) a variable or instance that is updated on every call to produce a
        #     result;
        # (2) an input variable or instance;
        # (3) both.

        # The method returns p_context, modified or not by evaluation of
        # p_expression on every matching object.

        # When you need to perform an action or computation on a lot of objects,
        # use this method instead of doing things like
           
        #              "for o in self.search(MyClass,...)"
        
        # Initialize the context variable "ctx"
        ctx = context
        i = 0
        for iid in self.search(className, ids=True, sortBy=sortBy, \
                               secure=secure, **fields):
            # Stop if the maximum number of objects has been reached
            if (max is not None) and (i >= max): break
            # Get the object from its ID
            o = self.getObject(iid, logMissing=True)
            # Ignore dead IDs
            if o is None: continue
            i += 1
            # Evaluate the expression
            if expression: exec(expression)
        return ctx

    def reindex(self, **kwargs):
        '''(re-/un-)indexes this object in the catalog corresponding to its
           class. Returns True if at least one change has been actually
           performed in at least one index.'''
        if self.isTemp(): raise Exception(INDEX_TEMP)
        handler = self.H()
        database = handler.server.database
        return database.reindexObject(handler, self, **kwargs)

    def getCatalog(self, raiseError=False):
        '''Returns the catalog storing p_self (if indexed)'''
        r = self.H().dbConnection.root.catalogs.get(self.class_.name)
        if not r and raiseError:
            raise Exception(CATALOG_KO % self.class_.name)
        return r

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Sub-PXs
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # The header displays, for a given object, its title and breadcrumb
    pxHeader = Px('''
     <div var="breadcrumb=ui.Breadcrumb(o, popup)"
          if="breadcrumb.parts" class=":breadcrumb.css">
      <x if="breadcrumb.sup">::breadcrumb.sup</x>
      <x for="part in breadcrumb.parts"
         var2="link=not loop.part.last and part.view">
       <img if="not loop.part.first" src=":url('arrowsA.svg')"
            class="iconS" style="transform: rotate(270deg)"/>
       <!-- Display the title or a link to the object -->
       <span if="not link">::part.title</span>
       <!-- Display a link for parent objects -->
       <a if="link" href=":part.url">::part.title</a>
      </x>
      <x if="breadcrumb.sub">::breadcrumb.sub</x>
     </div>''')

    # The grouped fields to show on a given page
    pxFields = Px('''
     <table width=":table.width | '100%'">
      <tr for="field in grouped"><td>:field.pxRender</td></tr></table>''')

    # PX displaying the content of a single field
    pxField = Px('''
     <div style="padding: 30px"
          var="field=o.getField(req.name);
               layout='view'">:field.pxRender</div>''',
     template=Template.px, hook='content')

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Object controls
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    pxControls = Px('''
     <!-- Object phases and pages. pxButtons will be called and inserted into
          the main phase. -->

     <x var="isEdit=layout == 'edit';
             nav=req.nav or 'no';
             current=phases.currentPage;
             compact=popup or config.ui.compact;
             picto='pictoC' if compact else 'picto';
             mayAct=not isEdit and guard.mayAct(o)">:phases.view</x>''')

    # Standard, custom and workflow-related actions
    pxButtons = Px('''
     <x if="phase.name == 'main'">
      <!-- Standard actions -->

      <!-- Save -->
      <div if="isEdit and current.showSave" var2="label=_('object_save')">
       <div>
        <a class="clickable"
           onclick=":'submitAppyForm(%s,%s,%s)' % \
                      (q('save'), q(current.name), q('view'))">
         <img src=":url('save.svg')" class=":picto"/>
         <div style=":'display:%s' % config.ui.pageDisplay">::label</div>
        </a>
       </div>
      </div>

      <!-- Cancel -->
      <div if="isEdit and current.showCancel" var2="label=_('object_cancel')">
       <div>
        <a class="clickable" name="cancel"
           onclick=":'submitAppyForm(%s,%s,%s)' % \
                      (q('cancel'), q(current.name), q('view'))">
         <img src=":url('cancel.svg')" class=":picto"/>
         <div style=":'display:%s' % config.ui.pageDisplay">::label</div>
        </a>
       </div>
      </div>

      <!-- Delete -->
      <div if="not isEdit and not popup and guard.mayDelete(o) and \
               not o.Lock.isSet(o, user, page)" var2="label=_('object_delete')">
       <div>
        <a class="clickable" onclick=":'onDeleteObject(%s)' % q(o.url)">
         <img src=":url('delete.svg')" class=":picto"/>
         <div style=":'display:%s' % config.ui.pageDisplay">::label</div>
        </a>
       </div>
      </div>

      <!-- Workflow transitions -->
      <x if="mayAct and o.class_.showTransitions(o, layout)"
         var2="workflow=o.getWorkflow()">
       <x var="transitions=workflow.getTransitions(o)" if="transitions">
        <x>:workflow.pxTransitions</x></x>
      </x>
     </x>

     <!-- Fields on layout "buttons". Layout is defined as "view" and not
          "cell", because "cell" would render a too minimalistic widget. -->
     <x if="not isEdit"
        var2="fields=o.getFields('buttons', page=req.page or 'main');
              layout='view'">
      <!-- Call view and not pxRender to avoid having a table -->
      <div for="field in fields"
           var2="name=field.name; value=field.getValueIf(o, name, layout)">
       <div>:field.cell</div></div>
     </x>''')

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Objet creation method
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def create(self, _name, secure=False, raiseOnWrongAttribute=True,
               executeMethods=True, initialComment=None, initialState=None,
               **kwargs):
        '''Create a new instance of a Appy class'''
        # If p__name is the name of a field, the created object will be linked
        # to p_self via this field. Else, p__name must correspond to a root
        # class and no link will exist between the created instance and p_self.

        # p_kwargs allow to specify values for object fields. If p_secure is
        # False, security checks will not be performed.

        # If p_raiseOnWrongAttribute is True, if a value from p_kwargs does not
        # correspond to a field on the created object, an AttributeError will be
        # raised. Else, the value will be silently ignored.

        # If p_executeMethods is False, the class's onEdit method, if present,
        # will not be called; any other defined method will not be called
        # neither (ie, Ref.insert, Ref.beforeLink, Ref.afterLink...).

        # p_initialComment will be stored as comment in the initial workflow
        # transition. If p_initialState is given (as a string), it will force
        # the object state instead of setting him to its workflow's initial
        # state.

        # Determine the ID of the object to create
        id = kwargs.pop('id') if 'id' in kwargs else None

        # Must we create an instance of a root class or one via a Ref field ?
        if _name[0].islower():
            # p__name is the name of a Ref field
            field = self.getField(_name)
            className = field.class_.meta.name
            # Check that the user can edit this field
            if secure: field.checkAdd(self)
        else:
            # p_name is the name of a root class
            className = _name
            field = None
        # Create the object
        handler = self.H()
        database = handler.server.database
        o = database.new(handler, className, id=id, secure=secure,
                       initialComment=initialComment, initialState=initialState)
        # Set object attributes
        for name, value in kwargs.items():
            try:
                setattr(o, name, value)
            except AttributeError as ae:
                if raiseOnWrongAttribute: raise ae
        # Call custom early initialization: a hook allowing the app developer to
        # initialise its object before it is linked to its initiator object.
        if executeMethods: multicall(o, 'onEditEarly', False, None)
        # Link the created object to its initiator when relevant
        if field: field.linkObject(self, o, executeMethods=executeMethods)
        # Call custom initialization: a hook allowing the app developer to
        # initialise its object after it's been linked to its initiator object.
        if executeMethods: multicall(o, 'onEdit', False, True)
        # Index the object (when relevant) and return it
        if o.class_.isIndexable():
            o.reindex()
        return o

    def getCopyValues(self, exclude=(), fields={}):
        '''Returns a dict ~{s_fieldName, fieldValue}~ representing p_self's
           dataset ready-to-be-copied to another object.'''
        # Check m_createFrom below to know more about method args
        # ~~~
        r = {}
        # Browse field values from p_self and store "copyable" ones in "r"
        for field in self.class_.fields.values():
            # Skip non persistent fields, back references & p_exclude(d) fields
            if not field.persist or (field.name in exclude) or \
               ((field.type == 'Ref') and field.isBack): continue
            # Use the value in p_fields if given
            if field.name in fields:
                value = fields[field.name]
            else:
                value = field.getCopyValue(self)
            if value is not None:
                r[field.name] = value
        return r

    def createFrom(self, _name, other, secure=False, executeMethods=True,
                   exclude=(), keepBase=False, keepHistory=False, fields={}):
        '''Similar to m_create above, excepted that another object (p_other) is
           used as base for filling in data for the object to create.'''
        # p_exclude lists fields (by their names) that will not be copied from
        # p_other. If p_keepBase is True, basic attributes will be kept on the
        # new object: creator and dates "created" and "modified". Else, the new
        # object's creator will be the logged user.

        # Note that this method does not perform a deep copy: objects linked via
        # Ref fields from p_self will be referenced by the clone, but not
        # themselves copied.

        # Get, as a dict, all field values to set from p_other. p_other may not
        # be of the same class as p_self.
        params = other.getCopyValues(exclude, fields)
        # Create the new instance
        r = self.create(_name, secure=secure, raiseOnWrongAttribute=False,
                        executeMethods=executeMethods, **params)
        # Copy p_other's history events as-is if requested
        if keepHistory: r.history.setEvents(other.history)
        # Propagate base attributes if required (which is useless if history has
        # already been cloned in its entirety).
        if not keepHistory and keepBase:
            history = r.history
            # Keep the same creator as p_other
            history[-1].login = other.creator
            # Keep the same creation date
            history[-1].date = other.created
            # Keep the same modification date
            history.modified = other.modified
        return r

    def delete(self, historize=False, executeMethods=True):
        '''Deletes p_self (see homonym method on the database object'''
        return self.H().server.database.delete(self, historize, executeMethods)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Translation method
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def translate(self, label, mapping=None, language=None, asText=False,
                  field=None, blankOnError=False):
        '''Translates a given p_label with p_mapping'''
        # If p_field is given, p_label does not correspond to a full label name,
        # but to a label type linked to p_field: "label", "descr" or "help".
        # Indeed, in this case, a specific i18n mapping may be available on the
        # field, so we must merge this mapping into p_mapping.
        handler = self.H()
        # In what language must we get the translation ?
        language = language or handler.guard.userLanguage
        # Get the label name, and the field-specific mapping if any
        if field:
            if field.type != 'group':
                fieldMapping = field.mapping[label]
                if fieldMapping:
                    if callable(fieldMapping):
                        fieldMapping = field.callMethod(self, fieldMapping)
                    if not mapping:
                        mapping = fieldMapping
                    else:
                        mapping.update(fieldMapping)
                # Translation may be found on a FieldTranslation instance
                if field.translations:
                    return field.translations.get(label, language, mapping)
            label = getattr(field, '%sId' % label)
        # Get the appropriate Translation object, or, if not found, get the one
        # corresponding to the defined fallback language.
        store = handler.dbConnection.root.objects
        translation = store.get(language) or \
                      store.get(self.config.ui.fallbackLanguage)
        # Get the label from this translation object, or from the fallback
        # language if not found.
        if translation:
            r = translation.get(label, mapping)
            if not r:
                # Try to get a translation in the fallback language
                fallback = self.config.ui.fallbackLanguage
                if fallback in store:
                    # The fallback language may not be among app's languages
                    r = store.get(fallback).get(label, mapping)
        else:
            r = ''
        # If still no result, put a nice name derived from the label instead of
        # a translated message.
        if not r and not blankOnError:
            return produceNiceMessage(label.rsplit('_', 1)[-1])
        # Manage carriage returns, depending on the desired format
        return r.replace('<br/>', '\n') if asText else r

    def callOnView(self):
        '''Call "onView" method if present'''
        method = getattr(self, 'onView', None)
        if method: method()

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # XML export
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    traverse['xml'] = 'perm:read'
    def xml(self):
        '''Returns p_self's XML version'''
        self.resp.setContentType('xml')
        return Marshaller(rootTag=self.class_.name).marshall(self, appy=True)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # PXs
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # The object, as shown in a list of referred (tied) objects
    pxTied = Px('''
     <tr var="ifield=initiator.field;
              *=ifield.initPx(initiator.o, req, _ctx_) if _ctx_.ajaxSingle \
                                                       else None;
              id=o.iid;
              objectIndex=ifield.getIndexOf(initiator.o, o, raiseError=False);
              mayView=guard.mayView(o);
              cbId='%s_%s' % (hook, currentNumber)"
         id=":id" valign=":ifield.rowAlign" class=":rowCss">
      <td for="column in columns" width=":column.width" align=":column.align"
          var2="field=column.field; isCB=field=='_checkboxes'"
          class=":selector.cbClass if isCB and selector else ''">

       <!-- Special column with the tied element's number -->
       <x if="(objectIndex!=None) and (field=='_number')">:ifield.pxNumber</x>

       <!-- Special column containing the tied element's checkbox -->
       <input if="isCB and mayView" type="checkbox" name=":hook" id=":cbId"
              var2="checked=cbChecked|False" checked=":checked"
              value=":id" onclick="toggleCb(this)"/>

       <!-- Any other column -->
       <x if="not column.special">:field.pxTied</x>
      </td>

      <!-- Store data in this tr node allowing to ajax-refresh it -->
      <script>:ifield.getAjaxDataRow(o, hook, rowCss=rowCss, \
               currentNumber=currentNumber, cbChecked=cbId, ajaxSingle=True, \
               initiator=':%d:%s' % (initiator.o.iid, ifield.name))</script>
     </tr>''')

    # The object, as shown in a list of search results
    pxResult = Px('''
     <tr valign=":uiSearch.search.rowAlign" class=":rowCss"
         var="id=o.iid; mayView=guard.mayView(o); backHook=str(id);
              cbId='%s_%s' % (mode.checkboxesId, currentNumber)"
         id=":id">
      <td for="column in mode.columns"
          var2="field=column.field; isCB=field=='_checkboxes'"
          id=":'field_%s' % field.name|field"
          width=":column.width" align=":column.align"
          class=":mode.cbClass if isCB else ''">

       <!-- Special column containing the checkbox -->
       <input if="isCB" type="checkbox" name=":mode.checkboxesId" id=":cbId"
              var2="checked=mode.cbChecked|True" value=":id"
              onclick="toggleCb(this)" checked=":checked"/>

       <!-- Any other column -->
       <x if="not column.special">:field.pxResult</x>
      </td>

      <!-- Store data in this tr node allowing to ajax-refresh it -->
      <script>:mode.getAjaxDataRow(o, rowCss=rowCss, \
               currentNumber=currentNumber, cbChecked=cbId)</script>
     </tr>''')

    # Base PX for viewing object fields from a given page
    view = Px('''
     <x var="x=guard.mayView(o, raiseError=True);
             table=o.getPageLayout(layout);
             page=o.getCurrentPage(req, layout);
             page,grouped,css,js,phases=o.getGroupedFields(page, layout)">
      <x>::ui.Includer.getSpecific(tool, css, js)</x>
      <x if="not req.notGlobal">::ui.Globals.getForms(tool)</x>
      <x var="tagId='pageLayout'; tagName=''; tagCss='';
              layoutTarget=o">:table.pxRender</x>
      <x if="not req.notGlobal"
         var="x=o.callOnView()">::ui.Globals.getScripts(tool, q, layout)</x>
     </x>''', template=Template.px, hook='content')

    # The default element computed and returned when an object is reached by a
    # traversal is the "view" PX.
    default = view

    # Base PX for editing object fields from a given page
    edit = Px('''
     <x var="x=guard.mayEdit(o, raiseError=True);
             table=o.getPageLayout(layout);
             page=o.getCurrentPage(req, layout);
             x=o.Lock.set(o, user, page);
             confirmText=req.confirmText;
             page,grouped,css,js,phases=o.getGroupedFields(page, layout);
             x=req.patchFromTemplate(o)">
      <x>::ui.Includer.getSpecific(tool, css, js)</x>
      <x>::ui.Globals.getForms(tool)</x>
      <script>:ui.Validator.getJsErrors(handler)</script>
      <!-- Warn the user that the form should be left via buttons -->
      <script>protectAppyForm()</script>
      <form id="appyForm" name="appyForm" method="post"
            enctype="multipart/form-data" action=":'%s/save' % o.url">
       <input type="hidden" name="action" value=""/>
       <input type="hidden" name="popup" value=":popup"/>
       <input type="hidden" name="page" value=":page"/>
       <input type="hidden" name="nav" value=":req.nav or ''"/>
       <input type="hidden" name="_get_" value=":req._get_ or ''"/>
       <input type="hidden" name="gotoPage" value=""/>
       <input type="hidden" name="gotoLayout" value=""/>
       <input type="hidden" name="insert" value=":req.insert or ''"/>
       <input type="hidden" name="confirmed" value="False"/>
       <x var="tagId='pageLayout'; tagName=''; tagCss='';
               layoutTarget=o">:table.pxRender</x>
      </form>
      <script if="confirmText">::'askConfirm(%s,%s,%s,null,%d)' % \
        (q('script'), q('postConfirmedEditForm()'), q(confirmText), \
         o.class_.getConfirmPopupWidth(popup))</script>
      <x>::ui.Globals.getScripts(tool, q, layout)</x>
     </x>''',

     js='''
      completeAppyForm = function(f) {
        // Complete Appy form p_f via special form element named "_get_"
        var g = f._get_;
        if (!g.value) return;
        var [source, name, info] = g.value.split(':');
        if (source == 'form') {
          // Complete p_f with form elements coming from another form
          var f2 = getNode(':' + name),
              fields = info.split(','),
              fieldName = fieldValue = null;
          if (!f2) return;
          for (var i=0; i<fields.length; i++) {
            fieldName = fields[i];
            if (fieldName[0] == '*') {
              // Try to retrieve search criteria from the storage if present
              fieldValue = sessionStorage.getItem(fieldName.substr(1));
              if (!fieldValue) continue;
              fieldName = 'criteria';
            }
            else {
              fieldValue = f2.elements[fieldName].value;
            }
            addFormField(f, fieldName, fieldValue);
          }
        }
      }
      submitAppyForm = function(action, gotoPage, gotoLayout) {
        var f = document.getElementById('appyForm');
        // Complete the form via the "_get_" element if present
        if (action != 'cancel') completeAppyForm(f);
        f.action.value = action;
        if (f.popup.value == 'True') {
          /* Initialize the "close popup" cookie. If set to "no", it is not time
             yet to close it. The timer hereafter will regularly check if the
             popup must be closed. */
          createCookie('closePopup', 'no');
          var popup = getNode('iframePopup', true);
          // Set a timer for checking when we must close the iframe popup
          popup.popupTimer = setInterval(backFromPopup, 700);
        }
        f.gotoPage.value = gotoPage;
        f.gotoLayout.value = gotoLayout;
        f.submit(); clickOn(f.elements[action]);
      }

      protectAppyForm = function() {
        window.onbeforeunload = function(e){
        var f = document.getElementById("appyForm");
        if (f.action.value == "") {
          var e = e || window.event;
          if (e) {e.returnValue = warn_leave_form;}
          return warn_leave_form;
          }
        }
      }''', template=Template.px, hook='content', name='edit')

    def getInitiator(self, search=False):
        '''Gets, from the request, information about a potential initiator'''
        # An initiator is an object from which an action is triggered on another
        # object (being p_self). The most obvious example is a target object
        # (p_self) created from a Ref field on a source object (the initiator).
        # The method returns None if there is no initiator.
        req = self.req
        nav = req.nav or req.onav
        if not nav or (nav == 'no'): return
        if not search:
            # Try to get a "classic" initiator, from the "[o]nav" key
            fieldType, info = nav.split('.', 1)
            field = eval(fieldType.capitalize())
            if field.initiator:
                r = field.initiator(self.tool, req, info)
                if r.isComplete():
                    return r
        else:
            # Try to get the initiator from the current search = a Ref field
            # referenced in key "_ref" or from search criteria.
            ref = req._ref or req.ref
            if not ref:
                criteria = Criteria.readFromRequest(self.H())
                if criteria and ('_ref' in criteria): ref = criteria['_ref']
            if ref:
                return self.getObject(ref.split(':')[0])

    def getFolder(self, create=False, withRelative=False):
        '''Gets, as a pathlib.Path instance, the folder where binary files
           related to this object are stored on the filesystem.'''
        return self.H().server.database.getFolder(self, create, withRelative)

    def cancel(self, initiator=None, popup=False, isTemp=False):
        '''Redirect the user after an object edition has been canceled'''
        self.H().commit = True
        # Render a message for the user. This message (and the url to redirect
        # the user to) may be customized by method m_onCancel.
        req = self.req
        message = url = None
        if hasattr(self, 'onCancel'):
            r = self.onCancel(isTemp)
            if r: message, url = r
        self.say(message or self.translate('object_canceled'))
        # Determine the current page
        page = req.page or 'main'
        if isTemp:
            # Delete the temp object whose creation has been canceled
            self.delete()
        else:
            # Remove the lock set on the current page
            Lock.remove(self, page)
        # If we are in a popup, render a minimalist HTML page that will close it
        if popup: return Iframe.goBack(self.tool, initiator)
        # Determine, if not defined yet, the URL to go back to
        if not url:
            if initiator:
                # Go back to the initiator page
                url = initiator.getUrl()
            elif isTemp:
                # Go back to home page
                url = self.tool.computeHomePage()
            else:
                # Return to p_self's view page
                url = self.getUrl(sub='view', page=page, nav=req.nav or 'no')
        self.goto(url)

    traverse['save'] = 'perm:write'
    def save(self):
        '''Called when p_self's data coming from the UI must be saved to disk.
           p_self can be a temporary object. In this case, "saving" consists in
           moving it to its "final" place.'''
        req = self.req
        tool = self.tool
        isTemp = self.isTemp()
        popup = req.popup == 'True'
        saveConfirmed = req.confirmed == 'True'
        # If this object is created from an initiator, get info about it
        initiator = self.getInitiator()
        # The precise action can be "save", but also "cancel"
        action = req.action
        if action == 'cancel': return self.cancel(initiator, popup, isTemp)
        # Create a validator: he will manage validation of request data
        validator = self.H().validator = Validator(self, saveConfirmed)
        r = validator.run()
        # If a result is returned, the "edit" page must be shown again:
        # validation failed or a confirmation is required.
        if r: return r
        # Update the object in the database
        text = self.H().server.database.update(self, validator, initiator)
        # Recompute cached info on the guard: things may have changed
        guard = self.guard
        guard.cache()
        # If p_self has already been deleted ("text" being None in that case),
        # or if the logged user has lost the right to view it, redirect him to
        # its home page.
        if (text is None) or not guard.mayView(self):
            if popup:
                if text: self.say(text)
                return Iframe.goBack(tool, initiator)
            else:
                return self.goto(tool.computeHomePage(), message=text)
        # If we are here, action is "save" and p_self has been updated. Redirect
        # the user to the appropriate page.
        self.say(text)
        # Come back from the popup if we were in it
        if popup and (req.gotoLayout == 'view'):
            return Iframe.goBack(tool, initiator)
        elif isTemp and initiator:
            # Let the initiator choose the back URL
            url = initiator.getBackUrl(self)
        else:
            # Go the the page as defined by request variables "gotoPage" and
            # "gotoLayout".
            url = self.getUrl(sub=req.gotoLayout, page=req.gotoPage,
                              nav=req.nav or 'no')
        return self.goto(url)

    traverse['new'] = True
    def new(self):
        '''Called when a user wants to create a root object or an object via a
           Ref field. A temporary object is created and the "edit" page to it is
           returned.'''
        req = self.req
        className = req.className
        if not className:
            self.log(MISS_CLS, type='error')
            self.raiseUnauthorized()
        # This requires a database commit
        handler = self.H()
        handler.commit = True
        # Create the params to add to the URL leading to the new object's edit
        # page.
        params = {'sub':'edit', 'page':'main', 'nav':'no',
                  'popup': req.popup == 'True'}
        initiator = self.getInitiator()
        if initiator:
            # Transmit initiator information in the next request (as well as
            # special key "_get_" if present). This information is stored in the
            # "nav" request key.
            params['nav'] = req.nav
            g = req._get_
            if g: params['_get_'] = g
            # Is the creation of the new object via the initiator allowed ?
            initiator.checkAllowed(log=True)
            # Let the initiator complete URL parameters when relevant
            initiator.updateParameters(params)
        # Create a temp object
        o = handler.server.database.new(handler, className, temp=True)
        # Populate p_o's edit form fields with a template object if present
        templateId = req.template
        if templateId: params['template'] = templateId
        # Call method "onCreate" if available. This method typically sets local
        # roles to the newly created object.
        if hasattr(o, 'onCreate'): o.onCreate()
        # Go to o's "edit" page
        self.goto(o.getUrl(**params))

    traverse['remove'] = 'perm:delete'
    def remove(self):
        '''Deletes p_self and return a message to the UI'''
        # Delete the object from the database. If a text is returned, an error
        # occurred and the deletion was aborted.
        text = self.delete(historize=True)
        # Prepare the back URL. None = the default = the referer page.
        back = None
        # Return a message to the UI
        handler = self.H()
        if not text:
            # The deletion has been successfully performed
            text = self.translate('action_done')
            # If we were consulting the object itself, return to the home page.
            # Else, return to the referer page (it is the default).
            if handler.isPublished(self):
                back = self.tool.computeHomePage()
        # If we are called from an Ajax request, simply return the message
        if handler.isAjax(): return text
        # Go to the "back" URL
        return self.goto(url=back, message=text)

    traverse['unlock'] = 'perm:write'
    def unlock(self):
        '''Unlocks the page whose name is in the request'''
        Lock.remove(self, page=self.req.page, force=True)
        self.goto(message=self.translate('action_done'))
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Workflow-specific model elements'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

# Default Appy permissions - - - - - - - - - - - - - - - - - - - - - - - - - - -
r, w, d = 'read', 'write', 'delete'
emptyDict = {}

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class WorkflowException(Exception): pass

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Role:
    '''A role is a prerogative that is granted to a user or group'''
    # A role can be local or global.
    # --------------------------------------------------------------------------
    # Local  | A local role applies to (and is stored on) a single object.
    # Global | A global role applies throughout the whole app, independently of
    #        | any object.
    # --------------------------------------------------------------------------
    # These roles are the base roles shipped with Appy
    baseRoles = (
     'Manager',      # The most permissive role. A Manager can do everything.
     'Owner',        # Automatically granted to an object's creator
     'Anonymous',    # Automatically granted to any unlogged hit to the site
     'Authenticated' # Automatically granted to any logged user
    )
    # The following roles are all global, "Owner" excepted
    baseLocalRoles = ('Owner',)
    # The following roles are "ungrantable": Appy grants them automatically
    baseUngrantableRoles = ('Anonymous', 'Authenticated')

    def __init__(self, name, local=False, grantable=True):
        # The name of the role
        self.name = name
        # Is it local or global ?
        self.local = local
        # It is a base Appy role or an application-specific one ?
        self.base = name in Role.baseRoles
        if self.base and (name in Role.baseLocalRoles):
            self.local = True
        # Is it grantable ?
        self.grantable = grantable
        if self.base and (name in self.baseUngrantableRoles):
            self.grantable = False

    def match(self, base, local, grantable):
        '''Does this role match parameters ?'''
        # Every parameter can be True, False or None
        if (base is not None) and (base != self.base): return
        if (local is not None) and (local != self.local): return
        if (grantable is not None) and (grantable != self.grantable): return
        return True

    def getLabel(self, withDefault=False):
        '''Returns the i18n label corresponding to this role. If p_withDefault
           is True, it also returns the default value for the label. In this
           case the r_esult is a tuple (s_label, s_defaultValue).'''
        r = 'role_%s' % self.name
        return r if not withDefault else r, self.name

    def __repr__(self):
        local = self.local and ' (local)' or ''
        return '<role %s%s>' % (self.name, local)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''This file contains basic classes that will be added into any user
   application for creating the basic structure of the application "Tool" which
   is the set of web pages used for configuring the application.'''

# ------------------------------------------------------------------------------
import types
import appy.gen as gen
from appy.fields import Field
from appy.fields.calendar import Calendar

# Prototypical instances of every type -----------------------------------------
class Protos:
    protos = {}
    # List of attributes that can't be given to a Type constructor
    notInit = ('id', 'type', 'pythonType', 'slaves', 'isSelect', 'hasLabel',
               'hasDescr', 'hasHelp', 'required', 'filterPx', 'validable',
               'isBack', 'pageName', 'masterName', 'select', 'sselect',
               'timeslots', 'legend', 'linkInPopup')
    @classmethod
    def get(self, appyType):
        '''Returns a prototype instance for p_appyType.'''
        className = appyType.__class__.__name__
        isString = (className == 'String')
        if isString:
            # For Strings, we create one prototype per format, because default
            # values may change according to format.
            className += str(appyType.format)
        if className in self.protos: return self.protos[className]
        # The prototype does not exist yet: create it
        if isString:
            proto = appyType.__class__(format=appyType.format)
            # Now, we fake to be able to detect default values
            proto.format = 0
        elif className == 'Pod':
            # Field "template" is mandatory
            proto = appyType.__class__(template='Appy.odt')
            proto.template = ()
        else:
            proto = appyType.__class__()
        self.protos[className] = proto
        return proto

# ------------------------------------------------------------------------------
class ModelClass:
    '''This class is the abstract class of all predefined application classes
       used in the Appy model: Tool, User, etc. All methods and attributes of
       those classes are part of the Appy machinery and are prefixed with _appy_
       in order to avoid name conflicts with user-defined parts of the
       application model.'''
    # In any ModelClass subclass we need to declare attributes in the following
    # list (including back attributes), to keep track of attributes order.
    _appy_attributes = []
    folder = False
    @classmethod
    def _appy_getTypeBody(klass, appyType, wrapperName):
        '''This method returns the code declaration for p_appyType'''
        typeArgs = ''
        proto = Protos.get(appyType)
        for name, value in appyType.__dict__.iteritems():
            # Some attrs can't be given to the constructor
            if name in Protos.notInit: continue
            # If the given value corresponds to the default value, don't give it
            if value == getattr(proto, name): continue
            if name == 'layouts':
                # For Tool attributes we do not copy layout info. Indeed, most
                # fields added to the Tool are config-related attributes whose
                # layouts must be standard.
                if klass.__name__ == 'Tool': continue
                layouts = appyType.getInputLayouts()
                # For the Translation class that has potentially thousands of
                # attributes, the most used layout is cached in a global var in
                # named "tfw" in wrappers.py.
                if (klass.__name__ == 'Translation') and \
                  (layouts=='{"edit":"f","search":"f","view":"f",}'):
                    value = 'tfw'
                else:
                    value = appyType.getInputLayouts()
            elif (name == 'klass') and value and (value == klass):
                # This is a auto-Ref (a Ref that references the klass itself).
                # At this time, we can't reference the class that is still being
                # defined. So we initialize it to None. The post-init of the
                # field must be done manually in wrappers.py.
                value = 'None'
            elif isinstance(value, basestring):
                value = '"%s"' % value
            elif isinstance(value, gen.Ref):
                if not value.isBack: continue
                value = klass._appy_getTypeBody(value, wrapperName)
            elif type(value) == type(ModelClass):
                moduleName = value.__module__
                if moduleName.startswith('appy.gen'):
                    value = value.__name__
                else:
                    value = '%s.%s' % (moduleName, value.__name__)
            elif isinstance(value, gen.Selection):
                value = 'Selection("%s")' % value.methodName
            elif isinstance(value, gen.Group):
                value = 'gps["%s"]' % value.name
            elif isinstance(value, gen.Page):
                value = 'pges["%s"]' % value.name
            elif callable(value):
                className = wrapperName
                if (appyType.type == 'Ref') and appyType.isBack:
                    className = value.im_class.__name__
                value = '%s.%s' % (className, value.__name__)
            typeArgs += '%s=%s,' % (name, value)
        return '%s(%s)' % (appyType.__class__.__name__, typeArgs)

    @classmethod
    def _appy_getBody(klass):
        '''This method returns the code declaration of this class. We will dump
           this in wrappers.py in the Zope product.'''
        className = klass.__name__
        # Determine the name of the class and its wrapper. Because so much
        # attributes can be generated on a TranslationWrapper, shortcutting it
        # to 'TW' may reduce the generated file from several kilobytes.
        if className == 'Translation': wrapperName = 'WT'
        else: wrapperName = 'W%s' % className
        res = 'class %s(%s):\n' % (className, wrapperName)
        # Tool must be folderish
        if klass.folder: res += '    folder=True\n'
        # First, scan all attributes, determine all used pages and groups, and
        # create a dict with it. It will prevent us from creating a new Page
        # instance for every field.
        pages = {}
        groups = {}
        layouts = []
        for name in klass._appy_attributes:
            exec 'field = klass.%s' % name
            if field.page.name not in pages:
                pages[field.page.name] = field.page
            group = field.group
            if group and (group.name not in groups):
                groups[group.name] = group
        # Generate the dict with pages
        res += '    pges={'
        for page in pages.itervalues():
            # Determine page "show" attributes
            pShow = ''
            for attr in ('',) + page.subElements:
                attrName = 'show%s' % attr.capitalize()
                pageShow = getattr(page, attrName)
                if isinstance(pageShow, basestring): pageShow='"%s"' % pageShow
                elif callable(pageShow):
                    pageShow = '%s.%s' % (wrapperName, pageShow.__name__)
                if pageShow != True:
                    pShow += ', %s=%s' % (attrName, pageShow)
            # For translation pages, fixed labels are used
            label = ''
            if className == 'Translation':
                name = (page.name == 'main') and 'Options' or page.name
                label = ', label="%s"' % name
            else:
                label = page.label and (', label="%s"' % str(page.label)) or ''
            res += '"%s":Pge("%s"%s%s),' % (page.name, page.name, pShow, label)
        res += '}\n'
        if groups:
            # Generate the dict with groups
            res += '    gps={'
            for group in groups.itervalues():
                label = group.label and (',label="%s"' % group.label) or ''
                res += '"%s":Grp("%s",style="%s"%s),' % \
                       (group.name, group.name, group.style, label)
            res += '}\n'
        # Secondly, dump every (not Ref.isBack) attribute
        for name in klass._appy_attributes:
            exec 'appyType = klass.%s' % name
            if (appyType.type == 'Ref') and appyType.isBack: continue
            typeBody = klass._appy_getTypeBody(appyType, wrapperName)
            res += '    %s=%s\n' % (name, typeBody)
        return res

# The User class ---------------------------------------------------------------
class User(ModelClass):
    _appy_attributes = ['password1', 'password2', 'title', 'name', 'firstName',
      'source', 'login', 'password3', 'password4', 'encrypted', 'email',
      'roles', 'resetPassword', 'groups', 'toTool', 'syncDate', 'lastLoginDate',
      'changePasswordAtNextLogin']
    # All methods defined below are fake. Real versions are in the wrapper.
    # Passwords are on a specific page.
    def showPassword12(self): pass
    def showPassword34(self): pass
    def showCancel12(self): pass
    def validatePassword(self): pass
    pp = {'page': gen.Page('passwords', showNext=False, show=showPassword12,
                          showCancel=showCancel12, label='User_page_passwords'),
          'width': 28, 'multiplicity': (1,1), 'format': gen.String.PASSWORD,
          'show': showPassword12, 'label': 'User'}
    password1 = gen.String(validator=validatePassword, **pp)
    password2 = gen.String(**pp)

    # Fields "password3" and "password4" are only shown when creating a user.
    # After user creation, those fields are not used anymore; fields "password1"
    # and "password2" above are then used to modify the password on a separate
    # page.
    pm = {'page': gen.Page('main', showPrevious=False, label='User_page_main'),
          'width': 28, 'layouts': Field.gLayouts, 'label': 'User',
          'group': gen.Group('main', style='grid', label='User_group')}
    title = gen.String(show='xml', indexed=True, **pm)
    def showName(self): pass
    name = gen.String(show=showName, **pm)
    firstName = gen.String(show=showName, **pm)
    pm['multiplicity'] = (1,1)
    def showLogin(self): pass
    def validateLogin(self): pass
    login = gen.String(show=showLogin, validator=validateLogin,
      indexed=True, **pm)
    del pm['label']
    password3 = gen.String(validator=validatePassword, show=showPassword34,
      format=gen.String.PASSWORD, label=('User', 'password1'), **pm)
    password4 = gen.String(show=showPassword34, format=gen.String.PASSWORD,
      label=('User', 'password2'), **pm)
    def getEncryptedPassword(self): pass
    def showEncryptedPassword(self): pass
    encrypted = gen.Computed(method=getEncryptedPassword,
                             show=showEncryptedPassword, layouts='f')
    def showEmail(self): pass
    pm['label'] = 'User'
    email = gen.String(show=showEmail, **pm)
    pm['multiplicity'] = (0, None)
    del pm['width']
    def showRoles(self): pass
    roles = gen.Select(show=showRoles, indexed=True, width=40, height=10,
      validator=gen.Selection('getGrantableRoles'), render='checkbox', **pm)

    # Where is this user stored? By default, in the ZODB. But the user can be
    # stored in an external LDAP (source='ldap').
    source = gen.String(show='xml', default='zodb', layouts='f')
    def doResetPassword(self): pass
    def showResetPassword(self): pass
    resetPassword = gen.Action(action=doResetPassword, show=showResetPassword,
                          confirm=True, label='User', icon='pwd', render='icon')

    # Some hidden fields
    hidden = {'show': False, 'layouts': 'f'}
    # For external users (source != "zodb"), we store the date of the last time
    # the external user and the local copy were synchronized.
    syncDate = gen.Date(format=gen.Date.WITH_HOUR, **hidden)
    # The date of the last login for this user
    lastLoginDate = gen.Date(format=gen.Date.WITH_HOUR, **hidden)
    # We may force a local user (source=zodb) to change its password at next
    # login (ie, for users created by an admin).
    changePasswordAtNextLogin = gen.Boolean(**hidden)

# The Group class --------------------------------------------------------------
class Group(ModelClass):
    _appy_attributes = ['title', 'login', 'roles', 'users', 'toTool2']
    # All methods defined below are fake. Real versions are in the wrapper.
    m = {'group': gen.Group('main', style='grid', label='Group_group'),
         'width': 25, 'indexed': True, 'layouts': Field.gLayouts,
         'label': 'Group'}
    title = gen.String(multiplicity=(1,1), **m)
    def showLogin(self): pass
    def validateLogin(self): pass
    login = gen.String(show=showLogin, validator=validateLogin,
                       multiplicity=(1,1), **m)
    roles = gen.String(validator=gen.Selection('getGrantableRoles'),
                       multiplicity=(0,None), **m)
    users = gen.Ref(User, multiplicity=(0,None), add=False, link='popup',
      height=15, back=gen.Ref(attribute='groups', show=User.showRoles,
                              multiplicity=(0,None), label='User'),
      showHeaders=True, shownInfo=('title', 'login', 'state*100px|'),
      actionsDisplay='inline', label='Group')

# The Translation class --------------------------------------------------------
class Translation(ModelClass):
    _appy_attributes = ['po', 'title', 'sourceLanguage', 'trToTool']
    # All methods defined below are fake. Real versions are in the wrapper.
    ta = {'label': 'Translation'}
    title = gen.String(show=False, indexed=True,
                     page=gen.Page('main', label='Translation_page_main'), **ta)
    def getPoFile(self): pass
    po = gen.Action(action=getPoFile, result='file', **ta)
    sourceLanguage = gen.String(width=4, **ta)
    def label(self): pass
    def show(self, name): pass

# The Page class ---------------------------------------------------------------
class Page(ModelClass):
    _appy_attributes = ['doc', 'title', 'content', 'pages', 'parent',
                        'expression', 'toTool3']
    folder = True
    pa = {'label': 'Page'}
    doc = gen.Pod(template='/gen/templates/Page.odt', formats=('pdf',),
                  show=False, layouts=gen.Table('f', width=None,
                                                css_class='inline', align=''))
    title = gen.String(show=('edit','xml'), multiplicity=(1,1),
                       indexed=True, **pa)
    content = gen.String(format=gen.String.XHTML, layouts='f')
    # If this Python expression returns False, the page can't be viewed
    def showExpression(self): pass
    expression = gen.String(layouts=Field.dLayouts, show=showExpression, **pa)
    # Pages can contain other pages
    def showSubPages(self): pass
    pages = gen.Ref(None, multiplicity=(0,None), add=True, link=False,
      composite=True, back=gen.Ref(attribute='parent', show=False, **pa),
      numbered=True, show=showSubPages, **pa)
Page.pages.klass = Page
setattr(Page, Page.pages.back.attribute, Page.pages.back)

# The Tool class ---------------------------------------------------------------
# Prefixes of the fields generated on the Tool
defaultToolFields = ('title', 'appyVersion', 'dateFormat', 'hourFormat',
                     'numberOfResultsPerPage', 'users', 'connectedUsers',
                     'synchronizeExternalUsers', 'groups', 'translations',
                     'loadTranslationsAtStartup', 'pages', 'calendar')

class Tool(ModelClass):
    # In a ModelClass we need to declare attributes in the following list
    _appy_attributes = list(defaultToolFields)
    folder = True

    # Tool attributes
    def forToolWriters(self): pass
    def pageForToolWriters(self): pass
    lf = {'layouts':'f'}
    title = gen.String(show=False, default='Configuration',
      page=gen.Page('main', show=False), **lf)
    appyVersion = gen.String(**lf)
    dateFormat = gen.String(default='%d/%m/%Y', **lf)
    hourFormat = gen.String(default='%H:%M', **lf)
    numberOfResultsPerPage = gen.Integer(default=30, **lf)

    # Ref(User) will maybe be transformed into Ref(CustomUserClass)
    ta = {'label': 'Tool'}
    userPage = gen.Page('users', show=pageForToolWriters,
                        label='Tool_page_users')
    users = gen.Ref(User, multiplicity=(0,None), add=True, link=False,
      composite=True, back=gen.Ref(attribute='toTool', show=False, layouts='f'),
      page=userPage, queryable=True, queryFields=('title', 'login', 'roles'),
      show=forToolWriters, showHeaders=True, actionsDisplay='inline',
      shownInfo=('title', 'name*15%|', 'login*20%|',
                 'roles*20%|', 'state*100px|'), **ta)

    def computeConnectedUsers(self): pass
    connectedUsers = gen.Computed(method=computeConnectedUsers, page=userPage,
                                  show='view', **ta)
    def doSynchronizeExternalUsers(self): pass
    def showSynchronizeUsers(self): pass
    synchronizeExternalUsers = gen.Action(action=doSynchronizeExternalUsers,
        show=showSynchronizeUsers, confirm=True, page=userPage, **ta)

    groups = gen.Ref(Group, multiplicity=(0,None), add=True, link=False,
      composite=True, back=gen.Ref(attribute='toTool2', show=False,layouts='f'),
      page=gen.Page('groups', show=pageForToolWriters,label='Tool_page_groups'),
      show=forToolWriters, queryable=True, queryFields=('title', 'login'),
      showHeaders=True, actionsDisplay='inline',
      shownInfo=('title', 'login*20%|', 'roles*20%|'), **ta)

    pt = gen.Page('translations', show=pageForToolWriters,
                  label='Tool_page_translations')
    translations = gen.Ref(Translation, multiplicity=(0,None), add=False,
      link=False, composite=True, show='view', page=pt, actionsDisplay='inline',
      back=gen.Ref(attribute='trToTool', show=False, layouts='f'), **ta)
    loadTranslationsAtStartup = gen.Boolean(default=True, show=False, page=pt,
                                            layouts='f')
    pages = gen.Ref(Page, multiplicity=(0,None), add=True, link=False,
      composite=True, show=('view', 'xml'), actionsDisplay='inline',
      back=gen.Ref(attribute='toTool3', show=False, layouts='f'),
      page=gen.Page('pages', show=pageForToolWriters, label='Tool_page_pages'),
      **ta)

    # This calendar is used for searches' "calendar" mode
    def calendarPreCompute(self, first, grid): pass
    def calendarAdditionalInfo(self, date, preComputed): pass
    calendar = Calendar(preCompute=calendarPreCompute,
                        additionalInfo=calendarAdditionalInfo, **ta)

    @classmethod
    def _appy_clean(klass):
        toClean = []
        for k, v in klass.__dict__.iteritems():
            if not k.startswith('__') and (not k.startswith('_appy_')):
                if k not in defaultToolFields:
                    toClean.append(k)
        for k in toClean:
            exec 'del klass.%s' % k
        klass._appy_attributes = list(defaultToolFields)
        klass.folder = True
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import re, difflib

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
innerDiff = re.compile('<span name="(insert|delete)".*? title="(.*?)">' \
                       '(.*?)</span>')
htmlTag = re.compile('<(?P<tag>\w+)( .*?)?>(.*)</(?P=tag)>')

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Merger:
    '''This class allows to merge 2 lines of text, each containing inserts and
       deletions.'''

    # Exception that may be raised by this class if the merge fails.
    class MergeError(Exception): pass

    def __init__(self, lineA, lineB, previousDiffs, differ):
        # lineA comes "naked": any diff previously found on it was removed from
        # it (ie, deleted text has been completely removed, while inserted text
        # has been included, but without its surrounding tag). Info about
        # previous diffs is kept in a separate variable "previousDiffs".
        self.lineA = lineA
        self.previousDiffs = previousDiffs
        # Differences between lineA and lineB have just been computed and are
        # included (within inner tags) in lineB. We will compute their position
        # in self.newDiffs (see below).
        self.lineB = lineB
        self.newDiffs = self.computeNewDiffs()
        # We choose to walk within self.lineB. We will keep in self.i our
        # current position within self.lineB.
        self.i = 0
        # The delta index that must be applied on previous diffs
        self.deltaPrevious = 0
        # A link to the caller HtmlDiff class
        self.differ = differ

    def computeNewDiffs(self):
        '''lineB may include inner "insert" and/or tags. This function
           detects them.'''
        i = 0
        res = []
        while i < len(self.lineB):
            match = innerDiff.search(self.lineB, i)
            if not match: break
            res.append(match)
            i = match.end()
        return res

    def getNextDiff(self):
        '''During the merging process on self.lineB, what next diff to
           "consume"? An old one? A new one?'''
        # No more diff ?
        if not self.previousDiffs and not self.newDiffs:
            return None, None, None
        # No more new diff ?
        if not self.newDiffs:
            diff = self.previousDiffs[0]
            del self.previousDiffs[0]
            return diff, diff.start() + self.deltaPrevious, True
        # No more previous diff ?
        if not self.previousDiffs:
            diff = self.newDiffs[0]
            del self.newDiffs[0]
            return diff, diff.start(), False
        # At least one more new and previous diff. Which one to consume?
        previousDiff = self.previousDiffs[0]
        newDiff = self.newDiffs[0]
        previousDiffIndex = previousDiff.start() + self.deltaPrevious
        newDiffIndex = newDiff.start()
        if previousDiffIndex <= newDiffIndex:
            # Previous wins
            del self.previousDiffs[0]
            return previousDiff, previousDiffIndex, True
        else:
            # New wins
            del self.newDiffs[0]
            return newDiff, newDiffIndex, False

    def manageBackOverlap(self, newDiff, oldText):
        '''p_newDiff has been removed from self.lineB. Here we check if there
           is no overlap with inserts from self.lineA, ie, text that was
           inserted in one if the many cumulative updates from self.lineA and
           that was deleted in self.lineB.'''
        # Before managing the overlap, check if there is one.
        oldDiff, oldDiffStart, isPrevious = self.getNextDiff()
        newDiffEnd = self.i + len(newDiff.group(3)) - len(oldText)
        if not oldDiff or not isPrevious or (oldDiffStart >= newDiffEnd):
            # There is no overlapping. Dump p_newDiff and the next diff as is
            # (if any).
            res = self.dumpNewDiff(newDiff)
            if oldDiff:
                # WARNING: oldDiffStart is not up-to-date! Indeed, we have
                # called getNextDiff (at the start of this method) BEFORE
                # calling dumpNewDiff (the line above). But dumpNewDiff updates
                # self.deltaPrevious. So we need to recompute oldDiffStart with
                # the current self.deltaPrevious
                if isPrevious:
                    oldDiffStart = oldDiff.start() + self.deltaPrevious
                res += self.dumpDiff(oldDiff, oldDiffStart, isPrevious)
            return res
        # If we are here, we must manage a back overlap. We will do it by
        # "consuming" p_newDiff.
        newText = newDiff.group(3)
        res = ''
        consumed = 0
        while True:
            # First, dump the part of p_newDiff that is not impacted by oldDiff.
            text = newText[consumed:oldDiffStart-self.i]
            if text:
                res += self.differ.getModifiedChunk(text, 'delete', '',
                                                    msg=newDiff.group(2))
                consumed += len(text)
            # Then, dump the part that overlaps with oldDiff
            text = oldDiff.group(3)
            res += self.differ.getModifiedChunk(text, 'delete', '',
                                                msg=newDiff.group(2))
            consumed += len(text)
            if consumed >= len(newText): break
            # Get the next diff
            oldDiff, oldDiffStart, isPrevious = self.getNextDiff()
            if not oldDiff or not isPrevious or (oldDiffStart > newDiffEnd):
                # End of the overlapping. Dump what remains in newText and dump
                # this next uncorrelated diff afterwards.
                res += self.differ.getModifiedChunk(newText[consumed:],
                                             'delete', '', msg=newDiff.group(2))
                self.i += len(newDiff.group(0))
                if oldDiff:
                    res += self.dumpDiff(oldDiff, oldDiffStart, isPrevious)
                return res
        # We have consumed p_newDiff entirely. Move forward within self.lineB
        # w.r.t p_newDiff.
        self.i += len(newDiff.group(0))
        return res

    def manageOverlap(self, oldDiff):
        '''p_oldDiff is a previously inserted text from self.lineA. This text
           is not found anymore at the start of self.lineB[self.i:]: it means
           that an overlapping diff exists among new diffs. We will manage this
           by identifying several, cutted, "insert" and/or "edit" zones.'''
        # The idea here is to "consume" the old inserted text until we have
        # found, within the new diff, all updates that have been performed on
        # this old text. Then, we will have found the complete "zone" that was
        # impacted by both old and new diffs.
        oldText = oldDiff.group(3)
        res = ''
        while oldText:
            # Get the overlapping (new) diff.
            newDiff, newDiffStart, isPrevious = self.getNextDiff()
            if not newDiff or (newDiffStart >= (self.i + len(oldText))):
                # No more new diff, or a new diff but far away, not within
                # oldText. So insert new the rest of p_oldText.
                # Invariant: at this point, we should find what remains in
                # oldText at self.lineB[self.i:].
                if not self.lineB[self.i:].startswith(oldText):
                    raise self.MergeError('An error occurred while computing ' \
                                          'overlapping diffs.')
                res += self.differ.getModifiedChunk(oldText, 'insert', '',
                                                    msg=oldDiff.group(2))
                self.i += len(oldText)
                oldText = ''
                # If we have "popped" a new diff, dump it anyway.
                if newDiff:
                    res += self.dumpDiff(newDiff, newDiffStart, isPrevious)
                break
            # Dump the part of the old text that has been untouched by the new
            # diff.
            if self.i < newDiffStart:
                untouched = self.lineB[self.i:newDiffStart]
                res += self.differ.getModifiedChunk(untouched, 'insert', '',
                                                    msg=oldDiff.group(2))
                self.i = newDiffStart
                oldText = oldText[len(untouched):]
            # Manage the new diff
            if (newDiff.group(1) == 'delete') and \
               len(newDiff.group(3)) > len(oldText):
                # Among deleted text, check if there is no overlap with previous
                # diffs (text deleted in self.lineB, might have been added in
                # one of the many cumulated updates in self.lineA).
                res += self.manageBackOverlap(newDiff, oldText)
                oldText = ''
            else:
                # Dump the new diff and update oldText
                res += self.dumpNewDiff(newDiff)
                if newDiff.group(1) == 'delete':
                    # Consume oldText, that was deleted, at least partly, by
                    # this diff.
                    oldText = oldText[len(newDiff.group(3)):]
        return res

    def dumpNewDiff(self, diff):
        '''Computes p_newDiff as it must appear in the result and return it.'''
        # Dump the new diff (from self.lineB)
        res = diff.group(0)
        # Move forward within self.lineB
        self.i += len(diff.group(0))
        # Because of this new diff, all indexes computed on self.lineA are now
        # wrong because we express them relative to lineB. So: update
        # self.deltaPrevious to take this into account.
        self.deltaPrevious += len(diff.group(0))
        if diff.group(1) == 'delete':
            # The indexes in self.lineA do not take the deleted text into
            # account, because it wasn't deleted at this time. So remove
            # from self.deltaPrevious the length of removed text.
            self.deltaPrevious -= len(diff.group(3))
        return res

    def dumpDiff(self, diff, diffStart, isPrevious):
        '''Computes the next p_diff (starting at p_diffStart) to insert into the
           result and return it. If p_isPrevious is True, the diff is an old one
           (from self.lineA); else, it is a new one (from self.lineB).'''
        # Dump the part of lineB between self.i and diffStart
        res = self.lineB[self.i:diffStart]
        self.i = diffStart
        if isPrevious:
            # Dump the old diff (from self.lineA)
            if diff.group(1) == 'insert':
                # Check if the inserted text is still present in lineB
                if self.lineB[self.i:].startswith(diff.group(3)):
                    # Yes. Dump the diff and go ahead within lineB
                    res += diff.group(0)
                    self.i += len(diff.group(3))
                else:
                    # The inserted text can't be found as is in lineB.
                    # Must have been (partly) re-edited or removed.
                    overlap = self.manageOverlap(diff)
                    res += overlap
            elif diff.group(1) == 'delete':
                res += diff.group(0)
        else:
            res += self.dumpNewDiff(diff)
        return res

    def merge(self):
        '''Merges self.previousDiffs into self.lineB.'''
        res = ''
        diff, diffStart, isPrevious = self.getNextDiff()
        while diff:
            res += self.dumpDiff(diff, diffStart, isPrevious)
            # Load the next diff, if any
            diff, diffStart, isPrevious = self.getNextDiff()
        # Dump the end of self.lineB if not completely consumed
        if self.i < len(self.lineB): res += self.lineB[self.i:]
        return res

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class HtmlDiff:
    '''This class allows to compute differences between two versions of some
       HTML chunk.'''

    # Default CSS styles to apply on a insertion or a deletion
    insertStyle = 'color:blue; cursor:help'
    deleteStyle = 'color:red; text-decoration:line-through; cursor:help'

    # Template tag that wraps a diff, be it a deletion or an insertion
    templateDiv = '<div style="%s" title="%s">%s</div>'

    def __init__(self, old, new,
                 insertMsg=u'Inserted text', deleteMsg=u'Deleted text',
                 insertCss=None, deleteCss=None, insertName='insert',
                 deleteName='delete', diffRatio=0.7):
        # p_old and p_new are strings containing chunks of HTML. If they are not
        # unicode strings, we convert them to unicode; this way, every char is
        # only one char lenght.
        self.old = old.strip()
        if isinstance(self.old, str): self.old = self.old.decode('utf-8')
        self.new = new.strip()
        if isinstance(self.new, str): self.new = self.new.decode('utf-8')
        # Every time an "insert" or "delete" difference will be detected from
        # p_old to p_new, the impacted chunk will be surrounded by a tag that
        # will get, respectively, a 'title' attribute filled p_insertMsg or
        # p_deleteMsg. The message will give an explanation about the change
        # (who made it and at what time, for example).
        self.insertMsg = insertMsg
        if isinstance(self.insertMsg, str):
            self.insertMsg = self.insertMsg.decode('utf-8')
        self.deleteMsg = deleteMsg
        if isinstance(self.deleteMsg, str):
            self.deleteMsg = self.deleteMsg.decode('utf-8')
        # This tag will get a CSS class p_insertCss or p_deleteCss for
        # highlighting the change. If no class is provided, default styles will
        # be used (see HtmlDiff.insertStyle and HtmlDiff.deleteStyle).
        self.insertCss = insertCss
        self.deleteCss = deleteCss
        # This tag will get a "name" attribute whose content will be
        # p_insertName or p_deleteName
        self.insertName = insertName
        self.deleteName = deleteName
        # The diff algorithm of this class will need to identify similarities
        # between strings. Similarity ratios will be computed by using method
        # difflib.SequenceMatcher.ratio (see m_isSimilar below). Strings whose
        # comparison will produce a ratio above p_diffRatio will be considered
        # as similar.
        self.diffRatio = diffRatio
        # Some computed values
        for tag in ('div', 'span'):
            for type in ('insert', 'delete'):
                setattr(self, '%s%sPrefix' % (tag, type.capitalize()),
                        '<%s name="%s"' % (tag, getattr(self, '%sName' % type)))

    def getModifiedChunk(self, seq, type, sep, msg=None):
        '''p_sep.join(p_seq) (if p_seq is a list) or p_seq (if p_seq is a
           string) is a chunk that was either inserted (p_type='insert') or
           deleted (p_type='delete'). This method will surround this part with
           a div or span tag that will get some CSS class allowing to highlight
           the update.

           If p_msg is given, it will be used instead of the default
           p_type-related message stored on p_self.'''
        # Will the surrounding tag be a div or a span ?
        tag = 'div' if sep == '\n' else 'span'
        # What message will it show in its 'title' attribute ?
        msg = msg or eval('self.%sMsg' % type)
        # What CSS class (or, if none, tag-specific style) will be used ?
        cssClass = eval('self.%sCss' % type)
        if cssClass:
            style = 'class="%s"' % cssClass
        else:
            style = eval('self.%sStyle' % type)
            style = 'style="%s"' % style
        # The 'name' attribute of the tag indicates the type of the update.
        tagName = eval('self.%sName' % type)
        # The idea is: if there are several lines, every line must be surrounded
        # by a tag. This way, we know that a surrounding tag can't span several
        # lines, which is a prerequisite for managing cumulative diffs.
        if sep == ' ':
            if not isinstance(seq, str):
                seq = sep.join(seq)
            sep = ''
        if isinstance(seq, str):
            return '%s<%s name="%s" %s title="%s">%s</%s>%s' % \
                   (sep, tag, tagName, style, msg, seq, tag, sep)
        else:
            r = ''
            for line in seq:
                r += '%s<%s name="%s" %s title="%s">%s</%s>%s' % \
                     (sep, tag, tagName, style, msg, line, tag, sep)
            return r

    def applyDiff(self, line, diff):
        '''p_diff is a regex containing an insert or delete that was found
           within line. This function applies the diff, removing or inserting
           the diff into p_line.'''
        # Keep content only for "insert" tags.
        content = ''
        if diff.group(1) == 'insert':
            content = diff.group(3)
        return line[:diff.start()] + content + line[diff.end():]

    def isSimilar(self, s1, s2):
        '''Returns True if strings p_s1 and p_s2 can be considered as
           similar.'''
        # Bypass the similarity algorithm for strings of length==1. Else, it can
        # lead to infinite loops between methods getHtmlDiff and getReplacement.
        if (len(s1) == 1) and (len(s2) == 1) and (s1 != s2): return False
        ratio = difflib.SequenceMatcher(a=s1.lower(), b=s2.lower()).ratio()
        return ratio > self.diffRatio

    def getLineAndType(self, line):
        '''p_line is a string that can already have been surrounded by an
           "insert" or "delete" tag. This is what we try to determine here.
           This method returns a tuple (type, line, innerDiffs, outerTag),
           where "type" can be:
           * "insert" if it has already been flagged as inserted;
           * "delete" if it has already been flagged as deleted;
           * None else;
           "line" holds the original parameter p_line, excepted:
           * if type="insert". In that case, the surrounding insert tag has been
             removed and placed into "outerTag" (a re.MatchObject from regex
             innerHtml, see above);
           * if inner diff tags (insert or delete) are found. In that case,
             - if inner "insert" tags are found, they are removed but their
               content is kept;
             - if inner "delete" tags are found, they are removed, content
               included;
             - "innerDiffs" holds the list of re.MatchObject instances
               representing the found inner tags.
        '''
        if line.startswith(self.divDeletePrefix):
            return ('delete', line, None, None)
        if line.startswith(self.divInsertPrefix):
            # Return the line without the surrounding tag
            action = 'insert'
            outerTag = htmlTag.match(line)
            line = outerTag.group(3)
        else:
            action = None
            outerTag = None
        # Replace found inner inserts with their content.
        innerDiffs = []
        while True:
            match = innerDiff.search(line)
            if not match: break
            # I found one.
            innerDiffs.append(match)
            line = self.applyDiff(line, match)
        return (action, line, innerDiffs, outerTag)

    def computeTag(self, regexTag, content):
        '''p_regexTag is a re.MatchObject from regex htmlTag. p_content is a
           new content to put within this tag. This method produces the new
           string tag filled with p_content.'''
        # Recompute start tag from p_regexTag
        startTag = '<%s' % regexTag.group(1)
        # Add tag attributes if found
        if regexTag.group(2):
            startTag += regexTag.group(2)
        startTag += '>'
        # Recompute end tag
        endTag = '</%s>' % regexTag.group(1)
        # Wrap content info reified tag
        return startTag + content + endTag

    def getSeqDiff(self, seqA, seqB, sep):
        '''p_seqA and p_seqB are lists of strings. Here we will try to identify
           similarities between strings from p_seqA and p_seqB, and return a
           list of differences between p_seqA and p_seqB, where each element
           is a tuple (action, line).
           * If p_action is "delete", "line" is a line of p_seqA considered as
             not included anymore in p_seqB;
           * If p_action is "insert", "line" is a line of p_seqB considered as
             not included in p_seqA;
           * If p_action is "replace", "line" is a tuple
             (lineA, lineB, previousDiffsA) containing one line from p_seqA and
             one from p_seqB considered as similar. "previousDiffsA" contains
             potential previous inner diffs that were found (but extracted
             from, for comparison purposes) lineA.
        '''
        res = []
        i = j = k = 0
        # Scan every string from p_seqA and try to find a similar string in
        # p_seqB.
        while i < len(seqA):
            pastAction, lineA, innerDiffs, outerTag=self.getLineAndType(seqA[i])
            if pastAction == 'delete':
                # We will consider this line as "equal" because it already has
                # been noted as deleted in a previous diff.
                res.append( ('equal', seqA[i]) )
            elif k == len(seqB):
                # We have already "consumed" every string from p_seqB. Remaining
                # strings from p_seqA must be considered as deleted (or
                # sometimes equal, see above)
                if not pastAction: res.append( ('delete', seqA[i]) )
                else:
                    # 'insert': should not happen. The inserted line should also
                    # be found in seqB.
                    res.append( ('equal', seqA[i]) )
            else:
                # Try to find a line in seqB which is similar to lineA.
                similarFound = False
                for j in range(k, len(seqB)):
                    if self.isSimilar(lineA, seqB[j]):
                        similarFound = True
                        # Strings between indices k and j in p_seqB must be
                        # considered as inserted, because no similar line exists
                        # in p_seqA.
                        if k < j:
                            for line in seqB[k:j]: res.append(('insert', line))
                        # Similar strings are appended in a 'replace' entry,
                        # excepted if lineA is already an insert from a
                        # previous diff: in this case, we keep the "old"
                        # version: the new one is the same, but for which we
                        # don't remember who updated it.
                        if (pastAction == 'insert') and (lineA == seqB[j]):
                            res.append( ('equal', seqA[i]) )
                        else:
                            res.append(('replace', (lineA, seqB[j],
                                                    innerDiffs, outerTag)))
                        k = j+1
                        break
                if not similarFound: res.append( ('delete', seqA[i]) )
            i += 1
        # Consider any "unconsumed" line from p_seqB as being inserted.
        if k < len(seqB):
            for line in seqB[k:]: res.append( ('insert', line) )
        # Merge similar diffs, excepted if separator is a carriage return
        if sep == '\n': return res
        newRes = []
        lastType = None
        for type, data in res:
            if lastType and (type != 'replace') and (lastType == type):
                newRes[-1] = (type, newRes[-1][1] + sep + data)
            else:
                newRes.append( (type, data) )
            lastType = type
        return newRes

    def split(self, s, sep):
        '''Splits string p_s with p_sep. If p_sep is a space, the split can't
           happen for a leading or trailing space, which must be considered as
           being part of the first or last word.'''
        # Manage sep == \n
        if sep == '\n': return s.split(sep)
        leadSpace = s.startswith(sep)
        trailSpace = s.endswith(sep)
        if not leadSpace and not trailSpace: return s.split(sep)
        res = s.strip(sep).split(sep)
        if leadSpace: res[0] = sep + res[0]
        if trailSpace: res[-1] = res[-1] + sep
        return res

    garbage = ('', '\r')
    def removeGarbage(self, l, sep):
        '''Removes from list p_l elements that have no interest, like blank
           strings or considered as is. Also: strip lines (ie, if sep is a
           carriage return.'''
        i = len(l)-1
        while i >= 0:
            if l[i] in self.garbage: del l[i]
            elif sep == '\n': l[i] = l[i].strip()
            i -= 1
        return l

    def getStringDiff(self, old, new):
        '''Identifies the differences between strings p_old and p_new by
           computing:
           * i = the end index of the potential common starting part (if no
                 common part is found, i=0);
           * jo = the start index in p_old of the potential common ending part;
           * jn = the start index in p_new of the potential common ending part.
        '''
        # Compute i
        i = -1
        diffFound = False
        while not diffFound:
            i += 1
            if (i == len(old)) or (i == len(new)): break
            if old[i] != new[i]: diffFound = True
        # i can't be inside an HTML tag.
        if (i > 0) and (old[i-1] == '<'): i -= 1
        # Compute jo and jn
        jo = len(old)
        jn = len(new)
        diffFound = False
        while not diffFound:
            if (jo == i) or (jn == i):
                # We have reached the end of substring old[i:] or new[i:]
                jo -=1
                jn -= 1
                break
            jo -= 1
            jn -= 1
            if old[jo] != new[jn]: diffFound=True
        return i, jo+1, jn+1

    def getDumpPrefix(self, res, add, previousAdd, sep):
        '''In most cases, when concatenating the next diff (p_add) to the
           global result (p_res), I must prefix it with p_sep (excepted if p_res
           is still empty). But when p_sep is a space, no space must be inserted
           between 2 adjacent updates (p_add and p_previousAdd), because such a
           space was not in the original version. This method computes the
           prefix, that can thus be empty if this latter case is met.'''
        prefix = ''
        if not res: return prefix
        if (sep == ' ') and previousAdd and \
           previousAdd.endswith('</span>') and add.startswith('<span'):
            pass
        else:
            prefix = sep
        return prefix

    def getReplacement(self, chunkA, chunkB, sep):
        '''p_chunkA has been replaced with p_chunkB. Compute this update and
           return it.'''
        res = ''
        # We know that some lines have been replaced from chunkA to chunkB. By
        # identifying similarities between those lines, consider some as having
        # been deleted, modified or inserted.
        previousAdd = None
        for action, line in self.getSeqDiff(chunkA, chunkB, sep):
            add = None
            if action in ('insert', 'delete'):
                add = self.getModifiedChunk(line, action, sep)
            elif action == 'equal':
                add = line
            elif action == 'replace':
                lineA, lineB, previousDiffsA, outerTagA = line
                # lineA has been replaced with lineB. Here, we will investigate
                # further here and explore differences at the *word* level
                # between lineA and lineB. previousDiffsA may contain a series
                # of updates (inserts, deletions) that have already been
                # performed on lineA. If lineA was a previously inserted line,
                # lineA comes without his outer tag, that lies in outerTagA
                # (as a re.MatchObject instance computed from regex htmlTag).
                # In that case, we will wrap the result with that tag.

                # As a preamble, and in order to restrict annoyances due to the
                # presence of XHTML tags, we will remove start and end tags
                # from lineA and lineB if present.
                i, ja, jb = self.getStringDiff(lineA, lineB)
                diff = self.getHtmlDiff(lineA[i:ja], lineB[i:jb], ' ')
                add = lineB[:i] + diff + lineB[jb:]
                # Merge potential previous inner diff tags that were found (but
                # extracted from) lineA.
                if previousDiffsA:
                    try:
                        merger = Merger(lineA, add, previousDiffsA, self)
                        add = merger.merge()
                    except Merger.MergeError:
                        # The merge algorithm has made a burn out. Simplify and
                        # consider lineA has having been completely deleted and
                        # lineB has completely inserted.
                        add = self.getModifiedChunk(lineA, 'delete', sep) + \
                              self.getModifiedChunk(lineB, 'insert', sep)
                # Rewrap line into outerTagA if lineA was a line tagged as
                # previously inserted.
                if outerTagA:
                    add = self.computeTag(outerTagA, add)
            if add: res += self.getDumpPrefix(res, add, previousAdd, sep) + add
            previousAdd = add
        return res

    def getHtmlDiff(self, old, new, sep):
        '''Returns the differences between p_old and p_new. Result is a string
           containing the comparison in HTML format. p_sep is used for turning
           p_old and p_new into sequences. If p_sep is a carriage return, this
           method is used for performing a whole diff between 2 strings splitted
           into sequences of lines; if sep is a space, the diff is a
           word-by-word comparison within 2 lines that have been detected as
           similar in a previous call to m_getHtmlDiff with sep=carriage
           return.'''
        res = ''
        a = self.split(old, sep)
        b = self.split(new, sep)
        matcher = difflib.SequenceMatcher()
        matcher.set_seqs(a, b)
        previousAdd = None
        for action, i1, i2, j1, j2 in matcher.get_opcodes():
            add = None
            # When sep is a space, we need to remember if we are dealing with
            # the last diff within the line or not.
            chunkA = self.removeGarbage(a[i1:i2], sep)
            chunkB = self.removeGarbage(b[j1:j2], sep)
            if action == 'equal':
                if chunkA: add = sep.join(chunkA)
            elif action == 'insert':
                if chunkB:
                    add = self.getModifiedChunk(chunkB, action, sep)
            elif action == 'delete':
                if chunkA:
                    add = self.getModifiedChunk(chunkA, action, sep)
            elif action == 'replace':
                if not chunkA and not chunkB:
                    pass
                elif not chunkA:
                    # Was an addition, not a replacement
                    add = self.getModifiedChunk(chunkB, 'insert', sep)
                elif not chunkB:
                    # Was a deletion, not a replacement
                    add = self.getModifiedChunk(chunkA, 'delete', sep)
                else: # At least, a true replacement
                    if (sep == ' ') and (sep not in chunkA) and \
                       (sep not in chunkB):
                        # By going here, we avoid infinite loops that may occur
                        # between m_getHtmlDiff and m_getReplacement
                        # (called below).
                        add = self.getModifiedChunk(chunkA, 'delete', sep) + \
                              self.getModifiedChunk(chunkB, 'insert', sep)
                    else:
                        add = self.getReplacement(chunkA, chunkB, sep)
            if add: res += self.getDumpPrefix(res, add, previousAdd, sep) + add
            previousAdd = add
        return res

    def get(self):
        '''Produces the result'''
        # Normally, if self.old is empty, the whole self.new should be
        # considered as inserted text. But see this line:
        if not self.old or not self.old.strip(): return self.new
        # Why? This is for avoiding problems in the case of cumulative diffs.
        # A cumulative diff means: calling HtmlDiff with, as old value, the
        # result of a previous call to HtmlDiff. In this case, if the whole text
        # is already considered as inserted, we will already have overlaps in
        # the next diff. Overlaps are hard to manage, so we avoid to get them
        # as a starting point when computing cumulative diffs.
        return self.getHtmlDiff(self.old, self.new, '\n')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Appy module managing log files'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import logging, sys, pathlib
from appy.model.utils import Object

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Logging-related parameters for your app'''
    # Available attributes to dump within log entries
    logAttributes = {
      'time': 'asctime', # The current date & time
      'level': 'levelname', # The log level
      'message': 'message' # The message to log, prefixed by the user login
    }

    def __init__(self, siteDateFormat='%Y/%m/%d %H:%M:%S',
                 appDateFormat='%Y/%m/%d %H:%M:%S',
                 siteAttributes=('time', 'message'),
                 appAttributes=('time', 'level', 'message'),
                 # Add "agent" hereafter to get the browser's User-Agent string
                 siteMessageParts=('ip', 'port', 'command', 'protocol',
                                   'path', 'message'),
                 appMessageParts=('user', 'message'),
                 siteSep=' | ', appSep=' | '):
        '''Initializes the logging configuration options.
           - p_siteDateFormat and p_appDateFormat define the format of dates
             dumped in log messages;
           - p_siteAttributes and p_appAttributes store the list of attributes
             that will be dumped in every log entry;
           - p_siteMessageParts and p_appMessageParts store the list of
             attributes contained within composite attribute "message";
           - p_siteSep and p_appSep store the separators that will be inserted
             between attributes.
        '''
        # Create a sub-object for splitting site- and app-related configuration
        # options.
        for type in ('app', 'site'):
            sub = Object(dateFormat=eval('%sDateFormat' % type),# ~pathlib.Path~
                         attributes=eval('%sAttributes' % type),
                         messageParts=eval('%sMessageParts' % type),
                         sep=eval('%sSep' % type))
            setattr(self, type, sub)

    def set(self, siteLogFolder, appLogFolder):
        '''Sets site-specific configuration elements'''
        # self.site.path is the path to the site log file, logging all HTTP
        # traffic on the site.
        self.site.path = pathlib.Path(siteLogFolder)
        # self.app.path is the path to the app-specific log, containing messages
        # dumped by the app and some by Appy itself.
        self.app.path = pathlib.Path(appLogFolder)
        # Typically, the site and app log files have standardized names and are
        # stored in <site>/var, with database-related files:
        # * "siteLogPath"    is <site>/var/site.log
        # * "appLogPath"     is <site>/var/app.log

    def getFormatter(self, type):
        '''Gets a logging.Formatter object for log entries of p_type'''
        sub = getattr(self, type)
        # Define the list of attributes to dump in every log entry
        attributes = []
        for name in sub.attributes:
            attributes.append('%%(%s)s' % Config.logAttributes[name])
        return logging.Formatter(sub.sep.join(attributes),
                                 datefmt=sub.dateFormat)

    def getLogger(self, type, debug=False):
        '''Return the site or app logger instance (depending on p_type), that
           will output log messages to self.siteLogPath or self.appLogPath. If
           p_debug is True, we are in debug mode: an additional handler will be
           defined for producing output on stdout.'''
        logger = logging.getLogger(type)
        # Get the path to the file where to log messages
        sub = getattr(self, type)
        path = sub.path
        # Add a file handler to the logger
        created = not path.is_file()
        path = str(path)
        logger.addHandler(logging.FileHandler(path))
        if debug:
            # Add an additional handler for outputing messages to stdout as well
            logger.addHandler(logging.StreamHandler(sys.stdout))
            level = logging.DEBUG
        else:
            level = logging.INFO
        # Messages under this level will not be dumped
        logger.setLevel(level)
        # Set a formatter for log entries
        formatter = self.getFormatter(type)
        for handler in logger.handlers:
            handler.setFormatter(formatter)
        # Return the created logger
        if created: logger.info('%s created.' % path)
        return logger
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

# ------------------------------------------------------------------------------
import re

from appy.px import Px
from appy.fields import Field
from appy.gen.layout import Layouts
from appy.shared import utils as sutils
from appy.shared.xml_parser import Escape
from appy.shared.xml_parser import XmlParser
from appy.fields.multilingual import Multilingual

# ------------------------------------------------------------------------------
class Replacements:
    '''Used by class Text2Html below, this class is used to define replacements
       for characters found in structured text, in order to recreate formatting
       from structured text written with some conventions.'''
    # Default replacement conventions
    default = {
      # Replace double quotes by "guillemets" (angle quotes). The sub-dict has
      # this meaning: if the previous char is a space (or if there is no
      # previous char), the double quote will be replaced with an opening
      # guillemet; else (special key "0"), it will be replaced by a closing
      # guillemet. The guillemet is flanked by a non-breakable space.
      '"': {' ': u'« ', '': u'« ', 0: u' »'},
      # Conventions for bold, italic and highlighting
      '[': '<b>', ']': '</b>',
      '<': '<i>', '>': '</i>',
      '{': '<span style="background-color: yellow">', '}': '</span>',
      '&': '&amp;' # XML escaping
    }
    # Among "default", some keys represent opening and ending delimiters
    delimiters = {'[': ']', '<': '>', '{': '}'}
    endDelimiters = sutils.flipDict(delimiters)
    # Prefix (semi)colons, question and exclamation marks with a non-breakable
    # space, excepted if such a space is already present.
    nbspChars = (':', ';', '!', '?', '%')
    for c in nbspChars:
        default[c] = {u' ': c, 0: u' %s' % c}
    whitespace = (' ', u' ')
    # Regular expression for replacing special markers (see "replacementFun"
    # attribute below.
    regex = re.compile('\*(.+?)\*')

    def __init__(self, replacements=None):
        # Apply custom replacements (if any). In order to disable a default
        # replacement, add it in p_replacement with a "None" value.
        if replacements:
            self.replacements = Replacements.default.copy()
            self.replacements.update(replacements)
        else:
            self.replacements = Replacements.default

    def getReplacementFor(self, char, previous):
        '''Return the replacement char for p_char. p_previous is the previous
           char encountered before p_char.'''
        r = self.replacements[char]
        if not isinstance(r, basestring):
            # The replacement depends on "previous"
            if previous in r:
                r = r[previous]
            else:
                r = r[0]
        return r

    def applyOn(self, text, fun=None, toReopen=None):
        '''Apply replacements on p_text and return a tuple
                           (s_result, [s_toReopen])
        '''
        # ----------------------------------------------------------------------
        #  s_result  | Is the resulting unicode string, whose special chars have
        #            | been replaced with their replacements.
        # ----------------------------------------------------------------------
        # [toReopen] | Among special chars to replace, there are opening and
        #            | ending chars, used as delimiters for applying style, ie,
        #            | bold and italic. After having processed p_text, for some
        #            | opening delimiters, we may not have encountered their
        #            | corresponding ending delimiter. If it occurs, in order to
        #            | produce a valid output, we will nevertheless dump, in
        #            | p_result, the replacements chars corresponding to the
        #            | missing ending delimiters. In that case, [toReopen] will
        #            | contain the list of these opening delimiters. This will
        #            | allow the caller be conscious of this fact, and, the next
        #            | time he will call us in order to dump a subsequent chunk
        #            | of p_text, he will possibly give us this list of tags
        #            | p_toReopen. Indeed, the real ending delimiters may be
        #            | present in the next p_text.
        # ----------------------------------------------------------------------
        # Ensure p_text is Unicode
        isStr = isinstance(text, str)
        if isStr: text = text.decode('utf-8')
        # Replace special markers when appropriate
        if fun: text = self.regex.sub(fun, text)
        # Build the result as a list
        r = []
        add = r.append
        repls = self.replacements
        delimiters = Replacements.delimiters
        endDelimiters = Replacements.endDelimiters
        previous = u'' # The previous word or whitespace char
        # Start by dumping p_toReopen delimiters if any
        stack = toReopen # The stack of currently opened delimiters
        if toReopen:
            for delimiter in toReopen:
                add(self.getReplacementFor(delimiter, None))
        # Dump "normal" p_text
        for char in text:
            if char not in repls:
                # Leave the char untouched in the result
                add(char)
            else:
                # Apply a replacement
                add(self.getReplacementFor(char, previous))
                # Update the stack
                if char in delimiters:
                    if stack is None: stack = []
                    stack.append(char)
                elif (char in endDelimiters) and stack:
                    # We suppose p_text to be well structured and the last
                    # opened delimiter to correspond to this end delimiter. But
                    # we protect ourselves against possible mess and avoid
                    # popping the stack if empty.
                    stack.pop()
            # Update "previous"
            if (char in self.whitespace) or (previous in self.whitespace):
                previous = char
            else:
                previous += char
        # If the stack is not empty, dump corresponding ending delimiters
        if stack:
            i = len(stack) - 1
            while i >= 0:
                add(self.getReplacementFor(delimiters[stack[i]], None))
                i -= 1
        # Join the result to a string
        r = u''.join(r)
        # Encode the result when relevant
        if isStr: r = r.encode('utf-8')
        return r, stack

class Text2Html:
    '''Converts a chunk of structured text into XHTML. Some conventions apply,
       ie, rows of text starting with a dash are converted to bulleted lists;
       rows starting with char "|" are considered to be table rows, etc.'''
    defaultReplacements = Replacements()

    def __init__(self, p='p', prefix='', replacements='default',
                 replacementsFun=None, preListClass=None, lastLiClass=None):
        # The HTML tag used for representing a paragraph
        self.p = p
        # A chunk of HTML code that could be inserted just after the first
        # opening tag in the result.
        self.prefix = prefix
        # The result, will be a list that'll be joined at the end of the process
        self.res = None
        # If p_replacements is:
        # ----------------------------------------------------------------------
        #   "default"    | A default Replacements instance will be used, in
        #                | order to format text according to some conventions
        #                | (see class hereabove)
        # ----------------------------------------------------------------------
        #   None         | Text replacement is disabled
        # ----------------------------------------------------------------------
        #   an instance  | A Replacements instance containing a set of
        #                | formatting options that may de different from the
        #                | defaults.
        # ----------------------------------------------------------------------
        if replacements == 'default':
            self.replacements = Text2Html.defaultReplacements
        else:
            self.replacements = replacements
        # If a p_fun(ction) is defined, structured p_text can contain special
        # markers of the form
        #                            *<name>*
        # Evertytime such marker is encountered in the text, p_fun will be
        # called, with "name" as unique arg, and must return the replacement
        # text for this name.
        self.replacementsFun = replacementsFun
        # If "preListClass" is specified, it corresponds to a CSS class that
        # will be applied to the paragraph that is positioned just before a list
        # (ol or ul).
        self.preListClass = preListClass
        # If "lastLiClass" is specified, it corresponds to a CSS class that will
        # be applied to the last bullet in a list.
        self.lastLiClass = lastLiClass
        # The list of opened delimiters not closed after the last text converted
        # with replacements (see attribute Replacements.delimiters hereabove).
        self.opened = None

    def add(self, part, suffix='\n'):
        '''Adds some p_part into the result'''
        res = self.res
        wasEmpty = not res
        res.append(part + suffix)
        if wasEmpty and self.prefix:
            # It is time to dump the prefix if specified
            res.append(self.prefix)

    def addTag(self, tag, start=True):
        '''Adds a (starting or ending, depending on p_start) tag to the
           result.'''
        if tag in ('ul', 'ol'):
            # Surround the list by a paragraph. Else, self.prefix could not be
            # rendered.
            if start:
                self.add('<%s>' % self.p, '')
                self.add('<%s>' % tag)
            else:
                self.add('</%s>' % tag, '')
                self.add('</%s>' % self.p)
        else:
            t = not start and '/' or ''
            self.add('<%s%s>' % (t, tag))

    def convertText(self, text):
        '''Converts, within p_text, some chars according to self.replacements
           and return the result. Use p_self.replacementsFun if defined and
           p_useReplacementFun is True.'''
        if not self.replacements: return text
        r, opened = self.replacements.applyOn(text, fun=self.replacementsFun,
                                              toReopen=self.opened)
        # "opened" may contain the list of opened delimiters found in p_text
        # that were not closed within p_text. Remember them in p_self.opened: we
        # will re-open them in the next call to p_converText, because they were
        # automatically closed, in p_text, by m_applyOn hereabove.
        self.opened = opened
        return r

    def applyPreListClass(self):
        '''Apply the CSS class in self.preListClass to the last paragraph in
           self.res.'''
        r = self.res
        if not r: return
        i = len(r) - 1
        while i >= 0:
            if r[i].startswith('<p>'):
                r[i] = '<p class="%s"%s' % (self.preListClass, r[i][2:])
                break
            i -= 1

    def isLast(self, rows, i, elem):
        '''Is row #p_i within p_rows the last to contain some p_elem ?'''
        return (i+1 == len(rows)) or not rows[i+1].startswith(elem)

    def convert(self, s):
        '''Returns the converted chunk of HTML'''
        if not s: return self.prefix
        self.res = []
        inList = False
        inTable = False
        i = 0
        rows = s.split('\n')
        total = len(rows)
        for row in rows:
            if row.startswith('- '):
                # Convert it to a bullet. But are we already in a list ?
                if not inList:
                    # Apply the "prelist" CSS class to the last paragraph
                    if self.preListClass: self.applyPreListClass()
                    self.addTag('ul')
                    inList = True
                # Is this the last bullet ?
                isLast = self.isLast(rows, i, '- ')
                if isLast:
                    start= '<li class="%s">' % self.lastLiClass
                    end = '</li>'
                    inList = False
                else:
                    start = '<li>'
                    end = '</li>'
                self.add('%s%s%s' % (start, self.convertText(row[2:]), end))
                # Add the end list tag when relevant
                if isLast: self.addTag('ul', start=False)
            elif row.startswith('|'):
                # Convert it to a table row. But are we already in a table ?
                if not inTable:
                    self.addTag('table')
                    inTable = True
                # Split the row in cells
                tr = '<tr>'
                tdh = 'td'
                for cell in row.strip('|').split('|'):
                    if cell.startswith('_'):
                        tdh = 'th'
                        cell = cell[1:]
                    tr += '<%s>%s</%s>' % (tdh, self.convertText(cell), tdh)
                tr += '</tr>'
                self.add(tr)
                if self.isLast(rows, i, '|'):
                    self.add('</table>')
                    self.inTable = False
            else:
                self.addTag(self.p)
                # Allow the use of dashes that remain dashes (and are not
                # converted to bullets, as above).
                if row.startswith('-'): row = '- %s' % row[1:]
                self.add('%s</%s>' % (self.convertText(row), self.p))
            i += 1
        # Close the last opened list if any
        if inList: self.addTag('ul', start=False)
        return ''.join(self.res)

class Html2Text(XmlParser):
    '''Converts a chunk HTML into structured text'''

    # This parser uses conventions for default replacements only
    # ~~~
    # Conversion of HTML start tags
    startTags = {'b':'[', 'strong':'[', 'i':'<', 'em':'<', 'li':'- '}

    # Conversion of HTML end tags
    endTags = {'b':']', 'strong':']', 'i':'>', 'em':'>', 'p':'\n', 'li':'\n'}

    def startDocument(self):
        XmlParser.startDocument(self)
        self.res = []

    def endDocument(self):
        self.res = ''.join(self.res).strip()
        return XmlParser.endDocument(self)

    def characters(self, content):
        content = content.replace('\n', '').replace('\t', '')
        self.res.append(content)

    def startElement(self, elem, attrs):
        '''Convert HTML start tag p_elem to its text counterpart'''
        token = Html2Text.startTags.get(elem)
        if token: self.res.append(token) # else, ignore the tag

    def endElement(self, elem):
        '''Convert HTML end tag p_elem to its text counterpart'''
        token = Html2Text.endTags.get(elem)
        if token: self.res.append(token) # else, ignore the tag

# ------------------------------------------------------------------------------
class Icon:
    '''An icon from the toolbar when the Text field is used in "structured"
       mode.'''

    def __init__(self, name, type,
                 label=None, icon=None, data=None, shortcut=None):
        # A short, unique name for the icon
        self.name = name
        # The following type of icons exist. Depending on the type, p_data
        # carries a specific type of information.
        # ----------------------------------------------------------------------
        # p_type      | p_data
        # ----------------------------------------------------------------------
        # "wrapper"   | the icon corresponds to a portion of text that will be
        #             | wrapped around a start and end char. p_data contains 2
        #             | chars: the start and end wrapper chars.
        #             | 
        #             | For example, icon "bold" is of type "wrapper", with data
        #             | being "[]". When applied to selected text "hello", it
        #             | becomes "[hello]".
        # ----------------------------------------------------------------------
        # "char"      | the icon corresponds to a char to insert into the field.
        #             | p_data is the char to insert.
        # ----------------------------------------------------------------------
        # "action"    | the icon corresponds to some action that is not
        #             | necessarily related to the field content. In that case,
        #             | p_data may be None or its sematincs may be specific to
        #             | the action.
        # ----------------------------------------------------------------------
        # "sentences" | a clic on the icon will display a menu containing
        #             | predefined sentences. Selecting one of them will inject
        #             | it in the target field, where the cursor is currently
        #             | set. In that case, p_data must hold the name of a
        #             | method that must exist on the current object. This
        #             | method will be called without arg and must return a list
        #             | of sentences, each one being a string.
        # ----------------------------------------------------------------------
        self.type = type
        # The i18n label for the icon's tooltip. Should include the keyboard
        # shortcut when present. If None, defaults to "icon_<name>"
        self.label = label or ('icon_%s' % name)
        # The name of the icon image on disk. If None, will be computed as
        # "icon_<name>.png".
        self.icon = icon or ('icon_%s' % name)
        # The data related to this icon, as described hereabove
        self.data = data
        # If a keyboard shortcut is tied to the icon, its key code is defined
        # here, as an integer. See JavasScript keycodes, https://keycode.info.
        self.shortcut = shortcut

    def asSentences(self, r, obj):
        '''For an icon of type "sentences", wraps the icon into a div allowing
           to hook the sub-div containing the sentences, and add this latter.'''
        # For an icon of type "sentences", add a div containing the sentences
        sentences = []
        tool = obj.getTool()
        for sentence in getattr(obj.appy(), self.data)():
            if not isinstance(sentence, basestring):
                # We have an additional, custom info to add besides the sentence
                # itself.
                sentence, info = sentence
            else:
                info = ''
            div = '<div class="sentence"><a class="clickable" ' \
                  'onclick="injectSentence(this)" title="%s">%s</a>%s</div>' % \
                  (sentence, tool.truncateValue(sentence, width=65), info)
            sentences.append(div)
        # Add a warning message if no sentence has been found
        if not sentences:
            sentences.append('<div class="legend">%s</div>' % \
                             obj.translate('no_sentence'))
        return '<div class="sentenceContainer" ' \
               'onmouseover="toggleDropdown(this) " ' \
               'onmouseout="toggleDropdown(this,\'none\')">%s' \
               '<div class="dropdown" style="display:none; width:350px">' \
               '%s</div></div>' % (r, '\n'.join(sentences))

    def get(self, obj):
        '''Returns the HTML chunk representing this icon'''
        shortcut = self.shortcut and str(self.shortcut) or ''
        tool = obj.getTool()
        r = '<img class="icon" src="%s" title="%s" name="%s"' \
            ' onmouseover="switchIconBack(this, true)"' \
            ' onmouseout="switchIconBack(this, false)"' \
            ' data-type="%s" data-data="%s" data-shortcut="%s" ' \
            'onclick="useIcon(this)"/>' % \
             (tool.getIncludeUrl(self.icon), tool.translate(self.label),
              self.name, self.type, self.data or '', shortcut)
        # Add specific stuff if icon type is "sentences"
        if self.type == 'sentences': r = self.asSentences(r, obj)
        return r

# All available icons
Icon.all = [Icon('bold',      'wrapper', data='[]',       shortcut=66),
            Icon('italic',    'wrapper', data='&lt;&gt;', shortcut=73),
            Icon('highlight', 'wrapper', data='{}',       shortcut=72),
            # Non breaking space
            Icon('blank',     'char',    data=' ',        shortcut=32),
            # Non breaking dash
            Icon('dash',      'char',    data='‑',        shortcut=54),
            # Increment the field height by <data>%
            Icon('lengthen',  'action',  data='30',       shortcut=56)]

# ------------------------------------------------------------------------------
class Text(Multilingual, Field):
    '''Field allowing to encode a text made of several paragraphs, implemented
       by a HTML textarea tag.'''
    # Make some classes available here
    Icon = Icon
    Replacements = Replacements
    ToHtml = Text2Html
    FromHtml = Html2Text

    # Default ways to render the field if multilingual
    defaultLanguagesLayouts = {
      'edit': 'horizontal', 'view': 'vertical', 'cell': 'vertical'}

    # Unilingual pxView
    pxViewUni = pxCellUni = Px('''
     <x>::field.getInlineEditableValue(obj, value or '-', layoutType, \
                                       name=name)</x>''')

    # The toolbar for structured content
    pxToolbar = Px('''
     <div class="toolbar" id=":tbid|field.name + '_tb'">
      <x for="icon in field.Icon.all">::icon.get(zobj)</x>
      <!-- Add inline-edition icons when relevant -->
      <x if="hostLayout">:field.pxInlineActions</x>
     </div>''',

     css = '''
      .toolbar { height: 24px; margin: 2px 0 }
      .sentenceContainer { position: relative; display: inline }
      .sentence { padding: 3px 0 }
      .icon { padding: 3px; border-width: 1px; border: 1px transparent solid }
      .iconSelected { background-color: #dbdbdb; border-color: #909090 }
     ''',

     js='''
      getIconsMapping = function(toolbar) {
        // Gets a mapping containing toolbar icons, keyed by their shortcut
        var r = {}, icons=toolbar.getElementsByClassName('icon');
        for (var i=0; i<icons.length; i++) {
          var icon=icons[i], key=icon.getAttribute('data-shortcut');
          if (key) r[parseInt(key)] = icon;
        }
        return r;
      }
      linkTextToolbar = function(toolbarId, target) {
        /* Link the toolbar with its target textarea. Get the target textarea if
           not given in p_target. */
        if (!target) {
          var targetId=_rsplit(toolbarId, '_', 2)[0];
          target = document.getElementById(targetId);
        }
        var toolbar=document.getElementById(toolbarId);
        toolbar['target'] = target;
        target['icons'] = getIconsMapping(toolbar);
      }
      switchIconBack = function(icon, selected) {
        icon.className = (selected)? 'icon iconSelected': 'icon';
      }
      lengthenArea = function(area, percentage) {
        // Lengthen some text p_area by some p_percentage
        var rate = 1 + (percentage / 100),
            styled = Boolean(area.style.height),
            height = (styled)? parseInt(area.style.height): area.rows;
        // Apply the rate
        height = Math.ceil(height * rate);
        // Reinject the new height to the correct area property
        if (styled) area.style.height = String(height) + 'px';
        else area.rows = height;
      }
      injectString = function(area, s) {
        // Inject some p_s(tring) into the text p_area, where the cursor is set
        var text = area.value,
                   start=area.selectionStart;
        area.value = text.substring(0, start) + s + \
                     text.substring(area.selectionEnd, area.value.length);
        area.selectionStart = area.selectionEnd = start +s.length;
        area.focus();
      }
      useIcon = function(icon) {
        // Get the linked textarea (if already linked)
        var area = icon.parentNode['target'];
        if (!area) return;
        var type=icon.getAttribute('data-type'),
            data=icon.getAttribute('data-data'),
            selectStart=area.selectionStart,
            selectEnd=area.selectionEnd,
            text=area.value;
        if (type == 'wrapper') {
          // Wrap the selected text within special chars
          area.value = text.substring(0, selectStart) + data[0] + \
                       text.substring(selectStart, selectEnd) + data[1] + \
                       text.substring(selectEnd, area.value.length);
          area.selectionStart = selectStart;
          area.selectionEnd = selectEnd + 2;
          area.focus();
        }
        else if (type == 'char') {
          // Insert a (sequence of) char(s) into the text
          injectString(area, data);
        }
        else if (type == 'action') {
          // Actions
          if (icon.name == 'lengthen') lengthenArea(area, parseInt(data));
        }
      }
      useShortcut = function(event, id) {
        if ((event.ctrlKey) && (event.keyCode in event.target['icons'])) {
          // Perform the icon's action
          useIcon(event.target['icons'][event.keyCode]);
          event.preventDefault();
        }
      }
      injectSentence = function(atag) {
        var area = atag.parentNode.parentNode.parentNode.parentNode['target'];
        if (!area) return;
        // Inject it
        injectString(area, atag.getAttribute('title'));
      }
     ''')

    # Buttons for saving or canceling while inline-editing the field. For a
    # structured text within a host layout, inline actions are rendered within
    # its toolbar.
    pxInlineActions = Px('''
      <div var="inToolbar=showToolbar and hostLayout;
                align=inToolbar and 'left' or 'right';
                fdir=inToolbar and 'row' or 'column'"
           style=":'float:%s;display:flex;flex-direction:%s' % (align, fdir)">
       <div><img id=":'%s_save' % tid" src=":url('save')"
                 class=":inToolbar and 'clickable icon' or 'inlineIcon'"
                 title=":_('object_save')"/></div>
       <div><img id=":'%s_cancel' % tid" src=":url('cancel')"
                 class=":inToolbar and 'clickable icon' or 'inlineIcon'"
                 title=":_('object_cancel')"/></div>
      </div>
      <script>:'prepareForAjaxSave(%s,%s,%s,%s)' % \
               (q(name),q(obj.id),q(obj.url),q(hostLayout))</script>''')

    # Unilingual pxEdit
    pxEditUni = Px('''
     <x var="tid=not lg and name or '%s_%s' % (name, lg);
             tbid='%s_tb' % tid;
             x=hostLayout and zobj.setLock(user, field=field);
             placeholder=field.getPlaceholder(obj);
             showToolbar=field.showToolbar(ignoreInner=hostLayout)">

      <!-- Show the toolbar when relevant -->
      <x if="showToolbar">:field.pxToolbar</x>

      <!-- Add buttons for inline-edition when relevant -->
      <x if="not showToolbar and hostLayout">:field.pxInlineActions</x>

      <!-- The text zone in itself -->
      <textarea id=":tid" name=":tid" cols=":field.getTextareaCols()"
       style=":field.getTextareaStyle()" placeholder=":placeholder"
       readonly=":field.isReadonly(zobj)" onkeydown=":field.onKeyDown(tid)"
       onfocus=":field.onFocus(tid, lg, hostLayout)"
       rows=":field.height">:field.getInputValue(inRequest, requestValue, value)
      </textarea></x>''')

    pxSearch = Px('''
     <input type="text" maxlength=":field.maxChars" size=":field.swidth"
            value=":field.sdefault" name=":widgetName"/>''')

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, indexed=False, mustIndex=True, indexValue=None, searchable=False,
      specificReadPermission=False, specificWritePermission=False, width=60,
      height=5, maxChars=None, colspan=1, master=None, masterValue=None,
      focus=False, historized=False, mapping=None, generateLabel=None,
      label=None, sdefault='', scolspan=1, swidth=None, sheight=None,
      persist=True, inlineEdit=False, view=None, cell=None, edit=None, xml=None,
      translations=None,
      # Specific attributes
      placeholder=None, languages=('en',), languagesLayouts=None,
      structured=False, readonly=False):
        # You can define a placeholder in the following attribute. Please
        # consult the homonym attribute on class String from string.py for more
        # information.
        self.placeholder = placeholder
        # If attribute "structured" is True, the text encoded by your users in
        # the field will implicitly follow some conventions for applying text
        # formatting. A toolbar will be shown and will offer shortcuts allowing
        # to inject formatting marks (or also special chars) into the textarea.
        self.structured = structured
        # If attribute "readonly" is True (or stores a method returning True),
        # the rendered textarea tag, on edit layouts, will have attribute
        # "readonly" set.
        self.readonly = readonly
        # Call the base constructors
        Multilingual.__init__(self, languages, languagesLayouts)
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          searchable, specificReadPermission, specificWritePermission, width,
          height, maxChars, colspan, master, masterValue, focus, historized,
          mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
          persist, inlineEdit, view, cell, edit, xml, translations)
        # Specify a filter PX if the field content is indexed
        if self.indexed: self.filterPx = 'pxFilterText'

    def getDefaultLayouts(self):
        '''Returns the default layouts for this field'''
        return self.inGrid() and Layouts.Text.g or Layouts.Text.b

    def getTextareaStyle(self):
        '''Get the content of textarea's "style" attribute'''
        # If the width is expressed as a string, the field width must be
        # expressed in attribute "style" (so returned by this method) and not
        # via attribute "cols" (returned by m_getTextareaCols below).
        return isinstance(self.width, str) and 'width:%s' % self.width or ''

    def getTextareaCols(self):
        '''Get the content of textarea's "cols" attribute'''
        # If width is expressed as an integer, it must be set as HTML attribute
        # "cols", and not as CSS attribute "width".
        return isinstance(self.width, int) and self.width or ''

    def showToolbar(self, ignoreInner=False):
        '''Show the toolbar if the field is structured and is not inner. In
           that latter case, the toolbar has already been rendered in the
           container field's headers.'''
        # Never show the tool bar if the field is not structured
        if not self.structured: return
        # Do not show the toolbar if the field is an inner field, provided this
        # check must be performed.
        return ignoreInner and True or not self.isInner()

    def onKeyDown(self, tid):
        '''When field is structured, this method returns the Javascript code to
           execute when a key is pressed, in order to implement keyboard
           shortcuts.'''
        if not self.structured: return ''
        return "useShortcut(event, '%s')" % tid

    def onFocus(self, tid, lg, hostLayout):
        '''When field is structured, this method returns the Javascript code to
           execute when the textarea gets focus, in order to link the field with
           the toolbar.'''
        if not self.structured: return ''
        if hostLayout:
            # We are inline-editing the (sub-)field: it has its own toolbar
            id = tid
        else:
            # For inner fields, there is a unique global toolbar
            id = lg and ('%s_%s' % (self.name, lg)) or self.name
        return "linkTextToolbar('%s_tb', this)" % id

    def getPlaceholder(self, obj):
        '''Returns a placeholder for the field if defined'''
        r = self.getAttribute(obj, 'placeholder') or ''
        if r == True:
            # A placeholder must be set, but we have no value. In this case, we
            # take the field label.
            r = obj.translate(self.labelId)
        return r

    def getUniFormattedValue(self, obj, value, layoutType='view',
          showChanges=False, userLanguage=None, language=None):
        '''If no p_language is specified, this method is called by
           m_getFormattedValue for getting a non-multilingual value (ie, in
           most cases). Else, this method returns a formatted value for the
           p_language-specific part of a multilingual value.'''
        if Field.isEmptyValue(self, obj, value) and not showChanges: return ''
        r = value
        if layoutType in ('view', 'cell'):
            r = Escape.xhtml(res, p=True)
        # If value starts with a carriage return, add a space; else, it will
        # be ignored.
        elif isinstance(r, basestring) and \
           (r.startswith('\n') or r.startswith('\r\n')): r = ' ' + r
        return r

    def getUniIndexValue(self, obj, value, forSearch):
        '''Gets the value to index for unilingual value p_value'''
        # The catalog does not like unicode strings
        if isinstance(value, unicode): value = value.encode('utf-8')
        return value

    def getUniStorableValue(self, obj, value):
        '''Gets the p_value as can be stored in the database within p_obj'''
        if not value: return value
        # Clean the value
        value = value.replace('\r', '')
        # Manage maxChars
        max = self.maxChars
        if max and (len(value) > max): value = value[:max]
        return value

    def getIndexType(self): return 'TextIndex'

    def getListHeader(self, ctx):
        '''If this field is structured, when used as an inner field, the
           toolbar must be rendered only once, within the container field's
           header row corresponding to this field.'''
        # Inject the toolbar when appropriate
        if (ctx['layoutType'] == 'edit') and self.showToolbar(ignoreInner=True):
            bar = self.pxToolbar(ctx)
        else:
            bar = ''
        return '%s%s' % (Field.getListHeader(self, ctx), bar.encode('utf-8'))
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Gridder:
    '''Specification about how to produce search results in grid mode. An
       instance of Gridder can be specified in static attribute
       SomeAppyClass.gridder.'''

    def __init__(self, width='350px', gap='20px', justifyContent='space-evenly',
                 alignContent='space-evenly', cols=None, margin=None,
                 showFields=True):
        # ~~~
        # If a p_width is specified, there will not be a fixed number of
        # columns: this number will depend on available space and this width.
        # If, for example, p_with is defined as being "350px" and the available
        # width is 800 pixels, the grid will have 2 columns.
        # ~~~
        # If p_cols is specified, p_width is ignored. p_cols defines the maximum
        # number of columns for the grid.
        # ~~~
        # The minimum width of every grid element
        self.width = width
        # The gap between grid elements. Will be passed to CSS property
        # "grid-gap".
        self.gap = gap
        # Specify how to align the whole grid horizontally inside the container.
        # Will be passed to CSS property "justify-content".
        self.justifyContent = justifyContent
        # Specify how to align the whole grid vertically inside the container.
        # Will be passed to CSS property "align-content".
        self.alignContent = alignContent
        # The maximum number of columns
        self.cols = cols
        # The gridder's margins
        self.margin = margin
        # By default, for every object rendered in a grid, the following
        # information is rendered, in that order:
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # a) | o.getSupTitle(nav)
        # b) | o.title
        # c) | o.getSubTitle()
        # d) | any other field as defined, either in an initiator Ref field
        #    | (via its attribute "shownInfo"), or via attribute "listColumns"
        #    | as defined on o's class.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # If p_showFields is False, element (d) will not be rendered.
        self.showFields = showFields

    def getContainerStyle(self):
        '''Returns the CSS styles that must be applied to the grid container
           element.'''
        # Determine columns
        if self.cols:
            columns = 'repeat(%d, 1fr)' % self.cols
        else:
            columns = 'repeat(auto-fill, minmax(%s, 1fr))' % self.width
        # Determine margins
        margin = '; margin:%s' % self.margin if self.margin else None
        
        # Return the CSS code
        return 'display:grid; grid-template-columns:%s; ' \
               'grid-gap:%s; justify-content:%s; align-content:%s%s' % \
               (columns, self.gap, self.justifyContent, self.alignContent,
                margin)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''A "request" stores HTTP GET request parameters and/or HTTP POST form
   values.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import re, urllib.parse

from appy.model.utils import Object

# Entries considered as empty in a multipart message - - - - - - - - - - - - - -
emptyMultiparts = (b'', b'--\r\n')

# Regular expression for parsing a Content-Disposition (cd) directive  - - - - -
cdRex = re.compile(b'name="([^"]+)"(?:.+filename="([^"]*)")?')

# Errors - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
EMPTY_POST = 'No Content-Length header for request with Content-Type "%s".'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Request(Object):
    '''Represents data coming from a HTTP request'''

    @classmethod
    def create(class_, handler):
        '''Analyses various elements from an incoming HTTP request (GET
           parameters, POST form data, cookies) and creates a Request object for
           storing them. If no element is found, an empty Request is created.'''
        # Create the Request instance
        req = Request()
        # Get parameters from a GET request if any
        qs = handler.parsedPath.query
        if qs: req.parse(qs)
        # Get POST form data if any
        contentType = handler.headers['Content-Type']
        if contentType:
            # Get the request length and content
            length = handler.headers['Content-Length']
            if length is None:
                handler.log('app', 'warning', EMPTY_POST % contentType)
                return req
            content = handler.rfile.read(int(length))
            # Several content encodings are possible
            if contentType == 'application/x-www-form-urlencoded':
                # Form elements are encoded in a way similar to GET parameters
                req.parse(content.decode('utf-8'))
            elif contentType.startswith('multipart/form-data'):
                # Form elements are encoded in a "multi-part" message, whose
                # elements are separated by some boundary text.
                boundary = '--' + contentType[contentType.index('boundary=')+9:]
                req.parseMultiPart(content, boundary.encode())
        # Get Cookies
        cookieString = handler.headers['Cookie']
        if cookieString is not None:
            # Cookies' (name, value) pairs are separated by semicolons
            req.parse(cookieString, sep=';')
        return req

    def addValue(self, name, value):
        '''Adds, on this request, an new attribute named p_name with this
           p_value. Manage multiple entries having the same p_name by creating a
           single attribute holding a list of values.'''
        # Ensure any string value is stripped
        name = name.strip()
        # Manage multiple values
        if name in self:
            # Update an existing value
            values = self[name]
            if isinstance(values, list):
                values.append(value)
            else:
                values = [values, value]
            value = values
        self[name] = value

    def parse(self, params, sep='&'):
        '''Parse GET/POST p_param(eters) and add one attribute per parameter'''
        for param in params.split(sep):
            # Every param is of the form <name>=<value> or simply <name>
            if '=' in param:
                # <name>=<value>
                name, value = param.split('=', 1)
                value = urllib.parse.unquote_plus(value)
            else:
                # <name>
                name = param
                value = None
            self.addValue(name, value)

    def parseMultiPart(self, content, boundary):
        '''Parses multi-part form content (bytes) and add to p_self one
           attribute per form element.'''
        for entry in content.split(boundary):
            # Ignore empty entries
            if entry in emptyMultiparts: continue
            # Separate attribute metadata from its value
            metadata, value = entry.split(b'\r\n\r\n', 1)
            # Search the various elements within metadata. For a standard
            # attribute, simply retrieve its name. For a file, also retrieve
            # its MIME type and file name.
            name = fileName = fileType = None
            for part in metadata.split(b'\r\n'):
                if not part: continue # Ignore empty chunks
                if name is None:
                    # "part" contains the directive "Content-Disposition" that
                    # includes the name of the attribute and, if the entry is a
                    # binary file, the name of the file.
                    name, fileName = cdRex.search(part).groups()
                    name = name.decode()
                    if fileName is not None: fileName = fileName.decode()
                elif fileType is None:
                    # "part" contains directive "Content-Type" containing the
                    # file's MIME type.
                    fileType = part.split()[1].decode()
            # Remove trailing space
            value = value[:-2]
            # For a (non empty) file, assemble the name, type and content in an
            # Object instance.
            if fileName and value:
                value = Object(name=fileName, type=fileType, value=value)
            else:
                # Get the value as a string
                value = value.decode()
            # Find the element value
            self.addValue(name, value)

    def patchFromTemplate(self, o):
        '''p_o is a temp object that is going to be edited via px "edit". If a
           template object must be used to pre-fill the web form, patch the
           request with the values retrieved on this template object.'''
        # This has sense only for temporary objects
        if not o.isTemp(): return
        id = self.template
        if not id: return
        template = o.getObject(id)
        # Some fields may be excluded from the copy
        toExclude = o.class_.getCreateExclude(o)
        # Browse fields
        for field in o.getFields('edit'):
            if field.name in toExclude: continue
            field.setRequestValue(template)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
